from __future__ import annotations

from typing import ClassVar

from pydantic import BaseModel, Field


class BookmarkArguments(BaseModel):
    BOOKMARK_ARGUMENTS_FROM_TUPLE_TO_DICT_VALIDATION_ERROR: ClassVar[
        str
    ] = "[BookmarkArguments] Failed to convert to dictionary because of the insufficient number of values."  # noqa: E501

    name: str = Field(min_length=1)
    prompt: str = Field(min_length=1)
    prepared_prompt: str = Field(min_length=1)
    response: str = Field(min_length=1)
    context_strategy: str
    prompt_template: str
    attack_module: str
    metric: str
    bookmark_time: str

    @classmethod
    def from_tuple_to_dict(cls, values: tuple) -> dict:
        """
        Creates a dictionary from a tuple of values representing BookmarkArguments fields.

        Args:
            values (tuple): A tuple containing the fields of BookmarkArguments in order.

        Returns:
            dict: A dictionary representing the BookmarkArguments.

        Raises:
            ValueError: If the number of values in the tuple is less than 10.
        """
        if len(values) < 10:
            raise ValueError(
                BookmarkArguments.BOOKMARK_ARGUMENTS_FROM_TUPLE_TO_DICT_VALIDATION_ERROR
            )

        return {
            "name": values[1],
            "prompt": values[2],
            "prepared_prompt": values[3],
            "response": values[4],
            "context_strategy": values[5],
            "prompt_template": values[6],
            "attack_module": values[7],
            "metric": values[8],
            "bookmark_time": values[9],
        }



import argparse
from ast import literal_eval

import cmd2
from rich.console import Console
from rich.table import Table

from moonshot.api import (
    api_delete_context_strategy,
    api_get_all_context_strategy_metadata,
    api_update_context_strategy,
)
from moonshot.integrations.cli.active_session_cfg import active_session
from moonshot.integrations.cli.utils.process_data import filter_data
from moonshot.src.redteaming.session.session import Session

console = Console()


def use_context_strategy(args: argparse.Namespace) -> None:
    """
    Use a context strategy for process the user's prompt (i.e. summarise past 3 prompts and add
    it to the current user's prompt)

    Args:
        args: A namespace with the context strategy parameters. Expected to have 'context_strategy'.
    """
    new_context_strategy_name = args.context_strategy
    num_of_prev_prompts = (
        args.num_of_prev_prompts
        if args.num_of_prev_prompts
        else Session.DEFAULT_CONTEXT_STRATEGY_PROMPT
    )

    # Check if current session exists. If it does, update context strategy and number of previous prompts
    if active_session:
        try:
            api_update_context_strategy(
                active_session["session_id"], new_context_strategy_name
            )
            active_session["context_strategy"] = new_context_strategy_name
            active_session["cs_num_of_prev_prompts"] = num_of_prev_prompts
            print(
                f"Updated session: {active_session['session_id']}. "
                f"Context Strategy: {active_session['context_strategy']}."
                f"No. of previous prompts for Context Strategy: {active_session['cs_num_of_prev_prompts']}."
            )
        except Exception as e:
            print(f"[use_context_strategy]: {str(e)}")

    else:
        print(
            "There is no active session. Activate a session to send a prompt with a context strategy."
        )


def list_context_strategies(args) -> list | None:
    """
    List all context strategies available.

    Args:
        args: A namespace object from argparse. It should have an optional attribute:
        find (str): Optional field to find context strategies with a keyword.
        pagination (str): Optional field to paginate context strategies.

    Returns:
        list | None: A list of ContextStrategy or None if there is no result.
    """
    try:
        context_strategy_metadata_list = api_get_all_context_strategy_metadata()
        keyword = args.find.lower() if args.find else ""
        pagination = literal_eval(args.pagination) if args.pagination else ()

        if context_strategy_metadata_list:
            filtered_context_strategies_list = filter_data(
                context_strategy_metadata_list, keyword, pagination
            )
            if filtered_context_strategies_list:
                _display_context_strategies(filtered_context_strategies_list)
                return filtered_context_strategies_list

        console.print("[red]There are no context strategies found.[/red]")
        return None

    except Exception as e:
        print(f"[list_context_strategies]: {str(e)}")


def clear_context_strategy() -> None:
    """
    Resets the context in a session.
    """
    # Check if current session exists
    if active_session:
        try:
            api_update_context_strategy(active_session["session_id"], "")
            active_session["context_strategy"] = ""
            print("Cleared context strategy.")
        except Exception as e:
            print(f"[clear_context_strategy: {str(e)}]")
    else:
        print(
            "There is no active session. Activate a session to send a prompt with a context strategy."
        )


def delete_context_strategy(args) -> None:
    """
    Deletes a context strategy after confirming with the user.

    Args:
        args (object): The arguments object. It should have a 'context_strategy' attribute
                       which is the ID of the context strategy to delete.
    """
    # Confirm with the user before deleting a context strategy
    confirmation = console.input(
        "[bold red]Are you sure you want to delete the context strategy (y/N)? [/]"
    )
    if confirmation.lower() != "y":
        console.print("[bold yellow]Context strategy deletion cancelled.[/]")
        return
    try:
        api_delete_context_strategy(args.context_strategy)
        print("[delete_context_strategy]: Context strategy deleted.")
    except Exception as e:
        print(f"[delete_context_strategy]: {str(e)}")


def _display_context_strategies(context_strategies: list) -> None:
    """
    Display a list of context strategies.

    This function takes a list of context strategies and displays them in a table format. If the list is empty,
    it prints a message indicating that no attack modules were found.

    Args:
        context_strategies (list): A list of context strategies.

    Returns:
        None
    """
    table = Table(
        title="Context Strategy List",
        show_lines=True,
        expand=True,
        header_style="bold",
    )
    table.add_column("No.", justify="left", width=2)
    table.add_column("Context Strategy Information", justify="left", width=98)
    for idx, context_strategy_data in enumerate(context_strategies, 1):
        context_strategy_data_str = ""
        for k, v in context_strategy_data.items():
            if k != "idx":
                context_strategy_data_str += f"[blue]{k.capitalize()}:[/blue] {v}\n\n"
        table.add_row(str(idx), context_strategy_data_str)
    console.print(table)


# Use context strategy arguments
use_context_strategy_args = cmd2.Cmd2ArgumentParser(
    description="Use a context strategy.",
    epilog="Example:\n use_context_strategy my_strategy_one",
)
use_context_strategy_args.add_argument(
    "context_strategy",
    type=str,
    help="The ID of the context strategy to use",
)
use_context_strategy_args.add_argument(
    "-n",
    "--num_of_prev_prompts",
    type=int,
    help="The number of previous prompts to use with the context strategy",
    nargs="?",
)

# Delete context strategy arguments
delete_context_strategy_args = cmd2.Cmd2ArgumentParser(
    description="Delete a context strategy.",
    epilog="Example:\n delete_context_strategy add_previous_prompt",
)

delete_context_strategy_args.add_argument(
    "context_strategy", type=str, help="The ID of the context strategy to delete"
)

# List context strategies arguments
list_context_strategies_args = cmd2.Cmd2ArgumentParser(
    description="List all context strategies.",
    epilog='Example:\n list_context_strategies -f "previous_prompt"',
)

list_context_strategies_args.add_argument(
    "-f",
    "--find",
    type=str,
    help="Optional field to find context strategies with keyword",
    nargs="?",
)

list_context_strategies_args.add_argument(
    "-p",
    "--pagination",
    type=str,
    help="Optional tuple to paginate context strategies(s). E.g. (2,10) returns 2nd page with 10 items in each page.",
    nargs="?",
)



from typing import Any, Optional

from pydantic import Field

from moonshot.src.datasets.dataset_arguments import (
    DatasetArguments as DatasetPydanticModel,
)


class CSV_Dataset_DTO(DatasetPydanticModel):
    id: Optional[str] = None  # Not required from user
    examples: Optional[Any] = None  # Not required from user
    name: str = Field(..., min_length=1)
    description: str = Field(default="", min_length=1)
    license: Optional[str] = ""
    reference: Optional[str] = ""
    file_path: str = Field(..., min_length=1)


class HF_Dataset_DTO(DatasetPydanticModel):
    id: Optional[str] = None  # Not a required from user
    examples: Optional[Any] = None  # Not a required from user
    name: str = Field(..., min_length=1)
    description: str = Field(default="", min_length=1)
    license: Optional[str] = ""
    reference: Optional[str] = ""
    params: dict = Field(..., min_length=1)



from dependency_injector.wiring import Provide, inject
from fastapi import APIRouter, Depends, HTTPException

from ..container import Container
from ..services.metric_service import MetricService
from ..services.utils.exceptions_handler import ServiceException

router = APIRouter(tags=["Metric"])


@router.get("/api/v1/metrics")
@inject
def get_all_metrics(
    metric_service: MetricService = Depends(Provide[Container.metric_service]),
) -> list[dict]:
    """
    Retrieve all metrics from the database.

    Args:
        metric_service (MetricService): The service responsible for retrieving metrics.

    Returns:
        list[str]: A list of all metrics.

    Raises:
        HTTPException: An error with status code 404 if no metrics are found.
                       An error with status code 400 if there is a validation error.
                       An error with status code 500 for any other server-side error.
    """
    try:
        return metric_service.get_all_metric()
    except ServiceException as e:
        if e.error_code == "FileNotFound":
            raise HTTPException(status_code=404, detail=e.msg)
        elif e.error_code == "ValidationError":
            raise HTTPException(status_code=400, detail=e.msg)
        else:
            raise HTTPException(status_code=500, detail=e.msg)


@router.delete("/api/v1/metrics/{metric_id}")
@inject
def delete_metric(
    metric_id: str,
    metric_service: MetricService = Depends(Provide[Container.metric_service]),
) -> dict[str, str] | tuple[dict[str, str], int]:
    """
    Delete a metric from the database by its ID.

    Args:
        metric_id (str): The unique identifier of the metric to delete.
        metric_service (MetricService): The service responsible for deleting the metric.

    Returns:
        dict[str, str] | tuple[dict[str, str], int]: A message indicating the successful deletion of the metric,
        or an HTTPException with an appropriate status code.

    Raises:
        HTTPException: An error with status code 404 if the metric is not found.
                       An error with status code 400 if there is a validation error.
                       An error with status code 500 for any other server-side error.
    """
    try:
        metric_service.delete_metric(metric_id)
        return {"message": "Metric deleted successfully"}
    except ServiceException as e:
        if e.error_code == "FileNotFound":
            raise HTTPException(
                status_code=404, detail=f"Failed to delete metric: {e.msg}"
            )
        elif e.error_code == "ValidationError":
            raise HTTPException(
                status_code=400, detail=f"Failed to delete metric: {e.msg}"
            )
        else:
            raise HTTPException(
                status_code=500, detail=f"Failed to delete metric: {e.msg}"
            )



from typing import Optional

from dependency_injector.wiring import Provide, inject
from fastapi import APIRouter, Depends, HTTPException

from ..container import Container
from ..schemas.endpoint_create_dto import EndpointCreateDTO, EndpointUpdateDTO
from ..schemas.endpoint_response_model import EndpointDataModel
from ..services.endpoint_service import EndpointService
from ..services.utils.exceptions_handler import ServiceException

router = APIRouter(tags=["Endpoint"])


@router.post("/api/v1/llm-endpoints")
@inject
def add_new_endpoint(
    endpoint_data: EndpointCreateDTO,
    endpoint_service: EndpointService = Depends(Provide[Container.endpoint_service]),
) -> dict[str, str] | tuple[dict[str, str], int]:
    """
    Add a new endpoint to the database.

    Args:
        endpoint_data (EndpointCreateDTO): The data transfer object containing endpoint details.
        endpoint_service (EndpointService): The service responsible for adding the endpoint.

    Returns:
        dict[str, str]: A message indicating the successful addition of the endpoint.

    Raises:
        HTTPException: An error with status code 404 if the file is not found.
                       An error with status code 400 if there is a validation error.
                       An error with status code 500 for any other server-side error.
    """
    try:
        endpoint_service.add_endpoint(endpoint_data)
        return {"message": "Endpoint added successfully"}
    except ServiceException as e:
        if e.error_code == "FileNotFound":
            raise HTTPException(
                status_code=404, detail=f"Failed to add endpoint: {e.msg}"
            )
        elif e.error_code == "ValidationError":
            raise HTTPException(
                status_code=400, detail=f"Failed to add endpoint: {e.msg}"
            )
        else:
            raise HTTPException(
                status_code=500, detail=f"Failed to add endpoint: {e.msg}"
            )


@router.get("/api/v1/llm-endpoints")
@inject
def get_all_endpoints(
    endpoint_service: EndpointService = Depends(Provide[Container.endpoint_service]),
) -> list[Optional[EndpointDataModel]]:
    """
    Get all the endpoints from the database.

    Args:
        endpoint_service (EndpointService): The service responsible for retrieving all endpoints.

    Returns:
        list[Optional[EndpointDataModel]]: A list of endpoint data models,
        which may contain None if an endpoint is not found.

    Raises:
        HTTPException: An error with status code 404 if the file is not found.
                       An error with status code 400 if there is a validation error.
                       An error with status code 500 for any other server-side error.
    """
    try:
        endpoint_models = endpoint_service.get_all_endpoints()
        for endpoint_model in endpoint_models:
            endpoint_model.mask_token()
        return endpoint_models
    except ServiceException as e:
        if e.error_code == "FileNotFound":
            raise HTTPException(
                status_code=404, detail=f"Failed to get endpoint: {e.msg}"
            )
        elif e.error_code == "ValidationError":
            raise HTTPException(
                status_code=400, detail=f"Failed to get endpoint: {e.msg}"
            )
        else:
            raise HTTPException(
                status_code=500, detail=f"Failed to get endpoint: {e.msg}"
            )


@router.get("/api/v1/llm-endpoints/name")
@inject
def get_all_endpoints_name(
    endpoint_service: EndpointService = Depends(Provide[Container.endpoint_service]),
) -> list[Optional[str]]:
    """
    Get all the endpoint names from the database.

    Args:
        endpoint_service (EndpointService): The service responsible for retrieving all endpoint names.

    Returns:
        list[Optional[str]]: A list of endpoint names, which may contain None if an endpoint name is not found.

    Raises:
        HTTPException: An error with status code 404 if the file is not found.
                       An error with status code 400 if there is a validation error.
                       An error with status code 500 for any other server-side error.
    """
    try:
        return endpoint_service.get_all_endpoints_names()
    except ServiceException as e:
        error_status_code = 500
        if e.error_code == "FileNotFound":
            error_status_code = 404
        elif e.error_code == "ValidationError":
            error_status_code = 400
        raise HTTPException(
            status_code=error_status_code,
            detail=f"Failed to get endpoint names: {e.msg}",
        )


@router.get("/api/v1/llm-endpoints/{endpoint_id}")
@inject
def get_endpoint(
    endpoint_id: str,
    endpoint_service: EndpointService = Depends(Provide[Container.endpoint_service]),
) -> EndpointDataModel | None:
    """
    Get an endpoint from the database by its ID.

    Args:
        endpoint_id (str): The unique identifier of the endpoint to retrieve.
        endpoint_service (EndpointService): The service responsible for retrieving the endpoint.

    Returns:
        EndpointDataModel | None: The endpoint data model if found, otherwise None.

    Raises:
        HTTPException: An error with status code 404 if the endpoint is not found.
                       An error with status code 400 if there is a validation error.
                       An error with status code 500 for any other server-side error.
    """
    try:
        endpoint_model = endpoint_service.get_endpoint(endpoint_id)
        if endpoint_model:
            endpoint_model.mask_token()
        return endpoint_model
    except ServiceException as e:
        if e.error_code == "FileNotFound":
            raise HTTPException(
                status_code=404, detail=f"Failed to get endpoint: {e.msg}"
            )
        elif e.error_code == "ValidationError":
            raise HTTPException(
                status_code=400, detail=f"Failed to get endpoint: {e.msg}"
            )
        else:
            raise HTTPException(
                status_code=500, detail=f"Failed to get endpoint: {e.msg}"
            )


@router.put("/api/v1/llm-endpoints/{endpoint_id}")
@inject
async def update_endpoint(
    endpoint_id: str,
    endpoint_data: EndpointUpdateDTO,
    endpoint_service: EndpointService = Depends(Provide[Container.endpoint_service]),
) -> dict[str, str] | tuple[dict[str, str], int]:
    """
    Update an existing endpoint in the database by its ID.

    Args:
        endpoint_id (str): The unique identifier of the endpoint to update.
        endpoint_data (EndpointCreateDTO): The data transfer object containing the updated endpoint details.
        endpoint_service (EndpointService): The service responsible for updating the endpoint.

    Returns:
        dict[str, str]: A message indicating the successful update of the endpoint.

    Raises:
        HTTPException: An error with status code 404 if the endpoint is not found.
                       An error with status code 400 if there is a validation error.
                       An error with status code 500 for any other server-side error.
    """
    try:
        endpoint_service.update_endpoint(endpoint_id, endpoint_data)
        return {"message": "Endpoint updated successfully"}
    except ServiceException as e:
        if e.error_code == "FileNotFound":
            raise HTTPException(
                status_code=404, detail=f"Failed to update endpoint: {e.msg}"
            )
        elif e.error_code == "ValidationError":
            raise HTTPException(
                status_code=400, detail=f"Failed to update endpoint: {e.msg}"
            )
        else:
            raise HTTPException(
                status_code=500, detail=f"Failed to update endpoint: {e.msg}"
            )


@router.delete("/api/v1/llm-endpoints/{endpoint_id}")
@inject
async def delete_endpoint(
    endpoint_id: str,
    endpoint_service: EndpointService = Depends(Provide[Container.endpoint_service]),
) -> dict[str, str] | tuple[dict[str, str], int]:
    """
    Delete an existing endpoint from the database by its ID.

    Args:
        endpoint_id (str): The unique identifier of the endpoint to delete.
        endpoint_service (EndpointService): The service responsible for deleting the endpoint.

    Returns:
        dict[str, str]: A message indicating the successful deletion of the endpoint.

    Raises:
        HTTPException: An error with status code 404 if the endpoint is not found.
                       An error with status code 400 if there is a validation error.
                       An error with status code 500 for any other server-side error.
    """
    try:
        endpoint_service.delete_endpoint(endpoint_id)
        return {"message": "Endpoint deleted successfully"}
    except ServiceException as e:
        if e.error_code == "FileNotFound":
            raise HTTPException(
                status_code=404, detail=f"Failed to delete endpoint: {e.msg}"
            )
        elif e.error_code == "ValidationError":
            raise HTTPException(
                status_code=400, detail=f"Failed to delete endpoint: {e.msg}"
            )
        else:
            raise HTTPException(
                status_code=500, detail=f"Failed to delete endpoint: {e.msg}"
            )


@router.get("/api/v1/connectors")
@inject
def get_all_connectors(
    endpoint_service: EndpointService = Depends(Provide[Container.endpoint_service]),
) -> list[Optional[str]]:
    """
    Get all the connector types from the database.

    Args:
        endpoint_service (EndpointService): The service responsible for retrieving all connector types.

    Returns:
        list[Optional[str]]: A list of connector types, which may contain None if a connector type is not found.

    Raises:
        HTTPException: An error with status code 404 if the file is not found.
                       An error with status code 400 if there is a validation error.
                       An error with status code 500 for any other server-side error.
    """
    # TODO - type check and model validation
    try:
        return endpoint_service.get_all_connectors()
    except ServiceException as e:
        if e.error_code == "FileNotFound":
            raise HTTPException(
                status_code=404, detail=f"Failed to get connector types: {e.msg}"
            )
        elif e.error_code == "ValidationError":
            raise HTTPException(
                status_code=400, detail=f"Failed to get connector types: {e.msg}"
            )
        else:
            raise HTTPException(
                status_code=500, detail=f"Failed to get connector types: {e.msg}"
            )



import argparse

import cmd2
from rich.console import Console

from moonshot.api import api_update_prompt_template
from moonshot.integrations.cli.active_session_cfg import active_session

console = Console()


def use_prompt_template(args: argparse.Namespace) -> None:
    """
    Use a prompt template by specifying its name while user is in a session.

    Args:
        args: A namespace with the prompt template parameters. Expected to have 'prompt_template'.
    """
    new_prompt_template_name = args.prompt_template

    # Check if current session exists
    if active_session:
        try:
            api_update_prompt_template(
                active_session["session_id"], new_prompt_template_name
            )
            active_session["prompt_template"] = new_prompt_template_name
            print(
                f"Updated session: {active_session['session_id']}. "
                f"Prompt Template: {active_session['prompt_template']}."
            )
        except Exception as e:
            print(f"[use_prompt_template]: {str(e)}")
    else:
        print(
            "There is no active session. Activate a session to send a prompt with a prompt template."
        )


def clear_prompt_template() -> None:
    """
    Resets the prompt template in a session.
    """
    # Check if current session exists
    if active_session:
        try:
            api_update_prompt_template(active_session["session_id"], "")
            active_session["prompt_template"] = ""
            print("Cleared prompt template.")
        except Exception as e:
            print(f"[clear_prompt_template: {str(e)}]")
    else:
        print(
            "There is no active session. Activate a session to send a prompt with a prompt template."
        )


# Use prompt template arguments
use_prompt_template_args = cmd2.Cmd2ArgumentParser(
    description="Use a prompt template.",
    epilog="Example:\n use_prompt_template 'analogical-similarity'",
)
use_prompt_template_args.add_argument(
    "prompt_template",
    type=str,
    help="Name of the prompt template",
)



import asyncio
import uuid
from typing import Any

from dependency_injector.wiring import inject

from moonshot.src.runners.runner import Runner

from ..services.auto_red_team_test_state import AutoRedTeamTestState
from ..services.base_service import BaseService
from ..services.runner_service import RunnerService
from ..status_updater.interface.redteam_progress_callback import (
    InterfaceRedTeamProgressCallback,
)
from ..types.types import RedTeamTestProgress


class AutoRedTeamTestManager(BaseService):
    @inject
    def __init__(
        self,
        auto_red_team_test_state: AutoRedTeamTestState,
        progress_status_updater: InterfaceRedTeamProgressCallback,
        runner_service: RunnerService,
    ) -> None:
        """
        Initialize the SessionService with dependencies.

        Args:
            progress_status_updater (InterfaceRedTeamProgressCallback): The callback interface for progress updates.
            runner_service (RunnerService): The service for managing runners.
        """
        self.auto_red_team_test_state = auto_red_team_test_state
        self.progress_status_updater = progress_status_updater
        self.runner_service = runner_service
        super().__init__()

    def generate_unique_task_id(self) -> str:
        unique_id = str(uuid.uuid4())
        return f"task_{unique_id}"

    def add_task(
        self, executor_id: str, task: asyncio.Task[Any], moonshot_runner: Runner
    ) -> None:
        self.auto_red_team_test_state.add_task(executor_id, task, moonshot_runner)

    def remove_task(self, executor_id: str) -> None:
        self.auto_red_team_test_state.remove_task(executor_id)

    async def cancel_task(self, executor_id: str) -> None:
        await self.auto_red_team_test_state.cancel_task(executor_id)

    def update_progress_status(self, updates: RedTeamTestProgress):
        self.auto_red_team_test_state.update_progress_status(updates)

    def on_task_completed(self, task: asyncio.Task[Any]) -> None:
        self.logger.debug(f"Task {task.get_name()} has completed")

    async def schedule_art_task(
        self, art_args: dict, active_runner: Runner, batch_size: int
    ) -> str:
        task_id = self.generate_unique_task_id()

        auto_redteam_coroutine = self.send_art_prompt(
            art_args, active_runner, batch_size
        )

        task = asyncio.create_task(auto_redteam_coroutine, name=task_id)

        def on_executor_completion(task: asyncio.Task[Any]):
            if task.exception():
                self.logger.error(
                    f"Executor {task.get_name()} has failed - {task.exception()}"
                )
            else:
                self.logger.debug(f"Executor {task.get_name()} has completed")

        task.add_done_callback(on_executor_completion)

        self.add_task(active_runner.id, task, active_runner)
        return active_runner.id

    async def send_art_prompt(self, art_args, active_runner: Runner, batch_size: int):
        await active_runner.run_red_teaming(
            {"attack_strategies": [art_args], "chat_batch_size": batch_size}
        )



from __future__ import annotations

from typing import Any, Callable

from pydantic import BaseModel, Field


class RunnerArguments(BaseModel):
    id: str  # The ID of the Runner.

    name: str = Field(min_length=1)  # The name of the Runner.

    database_file: str = ""  # The database file associated with the Runner.

    endpoints: list[str] = Field(min_length=1)  # List of endpoints for the Runner.

    description: str = ""  # A brief description of the Runner.

    # ------------------------------------------------------------------------------
    # These attributes are not exported to dict
    # ------------------------------------------------------------------------------
    database_instance: Any | None = (
        None  # The database instance associated with the Runner.
    )

    progress_callback_func: Callable | None = (
        None  # The progress callback function for the Runner.
    )

    def to_dict(self) -> dict:
        """
        Transforms the RunnerArguments instance into a dictionary format.

        This method serializes the RunnerArguments instance, excluding 'database_instance' and 'progress_callback_func',
        into a dictionary where attribute names become keys and their corresponding values are the dictionary values.

        Returns:
            A dictionary representation of the RunnerArguments instance, excluding non-serializable attributes.
        """
        return {
            "id": self.id,
            "name": self.name,
            "database_file": self.database_file,
            "endpoints": self.endpoints,
            "description": self.description,
        }



from pydantic import BaseModel


class ConnectorResponse(BaseModel):
    response: str = ""
    context: list = []

    def to_dict(self) -> dict:
        """
        Converts the ConnectorResponse instance to a dictionary.

        Returns:
            dict: A dictionary representation of the ConnectorResponse instance.
        """
        return {"response": self.response, "context": self.context}



import logging
import os
import sys
from logging.handlers import RotatingFileHandler
from typing import Literal

from dependency_injector import providers

from .types.types import UvicornLoggingConfig

COLORS = {
    "HEADER": "\033[95m",
    "OKBLUE": "\033[94m",
    "OKGREEN": "\033[92m",
    "WARNING": "\033[93m",
    "FAIL": "\033[91m",
    "ENDC": "\033[0m",
    "BOLD": "\033[1m",
    "UNDERLINE": "\033[4m",
    "WHITE": "\033[97m",
}


class ColorizedFormatter(logging.Formatter):
    LEVEL_COLORS = {
        logging.DEBUG: COLORS["OKBLUE"],
        logging.INFO: COLORS["OKGREEN"],
        logging.WARNING: COLORS["WARNING"],
        logging.ERROR: COLORS["FAIL"],
        logging.CRITICAL: COLORS["HEADER"],
    }

    def __init__(
        self,
        fmt: str,
        datefmt: str | None = None,
        style: Literal["%"] = "%",
        disableColor: bool = False,
    ):
        super().__init__(fmt, datefmt, style)
        self.disableColor = disableColor

    def format(self, record: logging.LogRecord):
        if self.disableColor:
            return super().format(record)
        else:
            color = str(self.LEVEL_COLORS.get(record.levelno))
            message = super().format(record)
            return color + message + COLORS["ENDC"]


def create_logging_dir(log_file_path: str):
    if not os.path.exists(log_file_path):
        os.makedirs(log_file_path)


def configure_app_logging(cfg: providers.Configuration):
    if cfg.log.logging():
        create_logging_dir(cfg.log.log_file_path())

    file_handler = RotatingFileHandler(
        filename=cfg.log.log_file_path() + "/web_api.log",
        maxBytes=cfg.log.log_file_max_size(),
        backupCount=cfg.log.log_file_backup_count(),
    )
    stream_handler = logging.StreamHandler(sys.stdout)
    stream_handler.setFormatter(ColorizedFormatter(cfg.log.format()))

    logging.basicConfig(
        handlers=[file_handler, stream_handler],
        level=cfg.log.level(),
        format=cfg.log.format(),
    )

    logging.info("Logging is configured.")


def create_uvicorn_log_config(cfg: providers.Configuration) -> UvicornLoggingConfig:
    if cfg.log.logging():
        create_logging_dir(cfg.log.log_file_path())

    return {
        "version": 1,
        "disable_existing_loggers": False,
        "formatters": {
            "default": {
                "()": "moonshot.integrations.web_api.logging_conf.ColorizedFormatter",
                "format": cfg.log.format(),
            },
            "file_formatter": {
                "()": "moonshot.integrations.web_api.logging_conf.ColorizedFormatter",
                "format": cfg.log.format(),
                "disableColor": True,
            },
        },
        "handlers": {
            "file": {
                "class": "logging.handlers.RotatingFileHandler",
                "filename": cfg.log.log_file_path() + "/web_api.log",
                "maxBytes": cfg.log.log_file_max_size(),
                "backupCount": cfg.log.log_file_backup_count(),
                "formatter": "file_formatter",
            },
            "console": {
                "class": "logging.StreamHandler",
                "stream": "ext://sys.stdout",
                "formatter": "default",
            },
        },
        "root": {
            "level": cfg.log.level(),
            "handlers": ["file", "console"],
        },
    }



from typing import Dict, List, Optional

from pydantic import BaseModel, Field


class ChatRecord(BaseModel):
    conn_id: str
    context_strategy: Optional[str]
    prompt_template: Optional[str]
    attack_module: Optional[str]
    metric: Optional[str]
    prompt: str
    prepared_prompt: str
    system_prompt: Optional[str]
    predicted_result: str
    duration: str
    prompt_time: str


class SessionMetadataModel(BaseModel):
    session_id: str = Field(min_length=1)
    description: Optional[str] = ""
    endpoints: List[str] = Field(min_length=1)
    created_epoch: str
    created_datetime: str = Field(min_length=1)
    prompt_template: Optional[str]
    context_strategy: Optional[str]
    cs_num_of_prev_prompts: int | None = 5
    attack_module: Optional[str]
    metric: Optional[str]
    system_prompt: Optional[str]


class SessionResponseModel(BaseModel):
    session_name: str
    session_description: str
    session: SessionMetadataModel
    chat_records: Optional[Dict[str, List[ChatRecord]]]



from __future__ import annotations

import asyncio
import time
from typing import Any, Callable

from moonshot.src.configs.env_variables import EnvVariables
from moonshot.src.results.result_arguments import ResultArguments
from moonshot.src.runners.runner_type import RunnerType
from moonshot.src.runs.run_arguments import RunArguments
from moonshot.src.runs.run_progress import RunProgress
from moonshot.src.runs.run_status import RunStatus
from moonshot.src.storage.db_interface import DBInterface
from moonshot.src.storage.storage import Storage
from moonshot.src.utils.import_modules import get_instance
from moonshot.src.utils.log import configure_logger

# Create a logger for this module
logger = configure_logger(__name__)


class Run:
    sql_table_name = "run_table"
    sql_create_run_table = """
        CREATE TABLE IF NOT EXISTS run_table (
        run_id INTEGER PRIMARY KEY AUTOINCREMENT,
        runner_id text NOT NULL,
        runner_type text NOT NULL,
        runner_args text NOT NULL,
        endpoints text NOT NULL,
        results_file text NOT NULL,
        start_time INTEGER NOT NULL,
        end_time INTEGER NOT NULL,
        duration INTEGER NOT NULL,
        error_messages text NOT NULL,
        raw_results text NOT NULL,
        results text NOT NULL,
        status text NOT NULL
        );
    """
    sql_create_run_record = """
        INSERT INTO run_table (
        runner_id,runner_type,runner_args,endpoints,results_file,start_time,end_time,duration,error_messages,raw_results,results,status)
        VALUES(?,?,?,?,?,?,?,?,?,?,?,?)
    """
    sql_read_run_record = """
        SELECT * from run_table WHERE run_id=?
    """
    sql_read_latest_run_record = """
        SELECT * FROM run_table WHERE run_id=(SELECT MAX(run_id) FROM run_table)
    """
    sql_read_all_run_records = """
        SELECT * FROM run_table
    """

    def __init__(
        self,
        runner_id: str,
        runner_type: RunnerType,
        runner_args: dict,
        database_instance: Any | None,
        endpoints: list[str],
        results_file: str,
        progress_callback_func: Callable | None = None,
    ) -> None:
        # Create run arguments
        self.run_arguments = RunArguments(
            runner_id=runner_id,
            runner_type=runner_type,
            runner_args=runner_args,
            database_instance=database_instance,
            endpoints=endpoints,
            results_file=results_file,
            start_time=0.0,
            end_time=0.0,
            duration=0,
            error_messages=[],
            raw_results={},
            results={},
            status=RunStatus.PENDING,
        )
        # Pass the reference of run_arguments to RunProgress
        self.run_progress = RunProgress(
            self.run_arguments,
            progress_callback_func,
        )
        # Create a cancellation asyncio event
        self.cancel_event = asyncio.Event()
        # Create run table
        if database_instance:
            Storage.create_database_table(database_instance, Run.sql_create_run_table)

    @staticmethod
    def load(database_instance: DBInterface | None, run_id: int | None) -> RunArguments:
        """
        Loads run data for a given run_id from the database, or the latest run if run_id is None.

        This method retrieves run data for the specified run_id from the database, or if run_id is None,
        it retrieves the latest run. If the database instance is not provided, it raises a RuntimeError.
        If the database instance is provided, it invokes the read_record method of the database instance
        with the given run_id or the latest run and returns a RunArguments object created from the retrieved record.

        Parameters:
            database_instance (DBAccessor | None): The database accessor instance.
            run_id (int | None): The ID of the run to retrieve, or None to retrieve the latest run.

        Returns:
            RunArguments: An object containing the details of the run with the given run_id or the latest run.
        """
        if not database_instance:
            raise RuntimeError("[Run] Database instance not provided.")

        if run_id is not None:
            run_arguments_info = Storage.read_database_record(
                database_instance,
                (run_id,),
                Run.sql_read_run_record,
            )
        else:
            run_arguments_info = Storage.read_database_record(
                database_instance,
                (),
                Run.sql_read_latest_run_record,
            )

        if run_arguments_info:
            return RunArguments.from_tuple(run_arguments_info)
        else:
            raise RuntimeError(
                f"[Run] Failed to get database record for run_id {run_id}: {database_instance}"
            )

    @staticmethod
    def get_all_runs(database_instance: DBInterface) -> list[RunArguments]:
        """
        Retrieves all run records from the database.

        This method fetches all the run records from the database and converts them into a list of RunArguments objects.
        If the database instance is not provided, it raises a RuntimeError. If no records are found, it also raises
        a RuntimeError.

        Parameters:
            database_instance (DBInterface): The database interface to fetch records from.

        Returns:
            list[RunArguments]: A list of RunArguments objects representing each run record.
        """
        if not database_instance:
            raise RuntimeError("[Run] Database instance not provided.")

        # Check that the table exists
        if not Storage.check_database_table_exists(
            database_instance, Run.sql_table_name
        ):
            return []

        all_run_arguments_info = Storage.read_database_records(
            database_instance,
            Run.sql_read_all_run_records,
        )

        if all_run_arguments_info:
            output = [RunArguments.from_tuple(info) for info in all_run_arguments_info]
            return output
        else:
            return []

    def cancel(self) -> None:
        """
        Sets the cancel event to stop the run process.

        This method is used to signal that the run process should be cancelled. It sets the cancel_event
        which can be checked in various points of the asynchronous run process to gracefully stop the execution.

        Returns:
            None
        """
        logger.warning("[Run] Cancelling run...")
        self.cancel_event.set()

    async def run(self) -> ResultArguments | None:
        """
        Executes the run process asynchronously.

        This method is the main entry point for running the process. It performs the run operation
        asynchronously and returns a ResultArguments object if the run completes successfully, or None
        if the run is cancelled or fails to complete.

        Returns:
            ResultArguments | None: The result of the run operation if successful, otherwise None.

        Raises:
            RuntimeError: If any error occurs during the run process.
        """
        # ------------------------------------------------------------------------------
        # Part 0: Initialise
        # ------------------------------------------------------------------------------
        logger.debug("[Run] Part 0: Initialising run...")
        start_time = time.perf_counter()
        try:
            # Initialise the run
            self.run_arguments.start_time = time.time()
            self.run_arguments.end_time = time.time()

            # Create a new run record in database
            if self.run_arguments.database_instance:
                inserted_record = Storage.create_database_record(
                    self.run_arguments.database_instance,
                    self.run_arguments.to_create_tuple(),
                    Run.sql_create_run_record,
                )
                if inserted_record:
                    self.run_arguments.run_id = inserted_record[0]
                else:
                    raise RuntimeError(
                        "[Run] Failed to create record: record not inserted."
                    )
            else:
                raise RuntimeError(
                    "[Run] Failed to create record: db_instance is not initialised."
                )

            # Set status to running
            self.run_progress.notify_progress(status=RunStatus.RUNNING)

        except Exception as e:
            error_message = (
                f"[Run] Failed to initialise run in Part 0 due to error: {str(e)}"
            )
            self.run_progress.notify_error(error_message)

        finally:
            logger.debug(
                f"[Run] Initialise run took {(time.perf_counter() - start_time):.4f}s"
            )

        # ------------------------------------------------------------------------------
        # Part 1: Get asyncio running loop
        # ------------------------------------------------------------------------------
        logger.debug("[Run] Part 1: Loading asyncio running loop...")
        loop = asyncio.get_running_loop()

        # ------------------------------------------------------------------------------
        # Part 2: Load runner and result processing module
        # ------------------------------------------------------------------------------
        logger.debug("[Run] Part 2: Loading modules...")
        start_time = time.perf_counter()
        runner_module_instance = None
        result_module_instance = None
        try:
            runner_module_instance = self._load_module(
                "runner_processing_module", EnvVariables.RUNNERS_MODULES.name
            )
            result_module_instance = self._load_module(
                "result_processing_module", EnvVariables.RESULTS_MODULES.name
            )
        except Exception as e:
            self.run_progress.notify_error(f"[Run] Module loading error: {e}")
        finally:
            logger.debug(
                f"[Run] Module loading took {(time.perf_counter() - start_time):.4f}s"
            )

        # ------------------------------------------------------------------------------
        # Part 3: Run runner processing module
        # ------------------------------------------------------------------------------
        logger.debug("[Run] Part 3: Running runner processing module...")
        start_time = time.perf_counter()
        runner_results = None
        try:
            if runner_module_instance:
                runner_results = await runner_module_instance.generate(  # type: ignore ; ducktyping
                    loop,
                    self.run_arguments.runner_args,
                    self.run_arguments.database_instance,
                    self.run_arguments.endpoints,
                    self.run_progress,
                    self.cancel_event,
                )
            else:
                raise RuntimeError("Failed to initialise runner module instance.")

        except Exception as e:
            error_message = f"[Run] Failed to run runner processing module in Part 3 due to error: {str(e)}"
            self.run_progress.notify_error(error_message)

        finally:
            logger.debug(
                f"[Run] Running runner processing module took {(time.perf_counter() - start_time):.4f}s"
            )

        # ------------------------------------------------------------------------------
        # Part 4: Run result processing module
        # ------------------------------------------------------------------------------
        logger.debug("[Run] Part 4: Running result processing module...")
        start_time = time.perf_counter()
        updated_runner_results = None
        try:
            if result_module_instance:
                updated_runner_results = result_module_instance.generate(  # type: ignore ; ducktyping
                    runner_results
                )
                if updated_runner_results:
                    self.run_progress.notify_progress(
                        results=updated_runner_results.results
                    )
            else:
                raise RuntimeError("Failed to initialise result module instance.")

        except Exception as e:
            error_message = f"[Run] Failed to run result processing module in Part 4 due to error: {str(e)}"
            self.run_progress.notify_error(error_message)

        finally:
            logger.debug(
                f"[Run] Running result processing module took {(time.perf_counter() - start_time):.4f}s"
            )

        # ------------------------------------------------------------------------------
        # Part 5: Wrap up run
        # ------------------------------------------------------------------------------
        logger.debug("[Run] Part 5: Wrap up run...")
        return updated_runner_results

    def _load_module(self, arg_key: str, env_var: str):
        """
        Load a module based on the argument key and environment variable.

        This method retrieves the module name from the runner arguments using the provided
        argument key. It then attempts to load the module using the name and the file path
        obtained from the environment variable. If the module name is not provided or the
        module instance cannot be created, a RuntimeError is raised.

        Args:
            arg_key (str): The key to look up the module name in the runner arguments.
            env_var (str): The environment variable used to obtain the file path of the module.

        Returns:
            An instance of the loaded module.

        Raises:
            RuntimeError: If the module name is not provided or the module instance cannot be created.
        """
        module_name = self.run_arguments.runner_args.get(arg_key)
        if not module_name:
            raise RuntimeError(f"[Run] Module name for '{arg_key}' not provided.")
        module_instance = get_instance(
            module_name,
            Storage.get_filepath(env_var, module_name, "py"),
        )
        if not module_instance:
            raise RuntimeError(
                f"[Run] Unable to get instance for module '{module_name}'."
            )
        return module_instance()



import importlib.resources
from enum import Enum
from pathlib import Path

from moonshot.src.utils.log import configure_logger

# Create a logger for this module
logger = configure_logger(__name__)

__app_name__ = "moonshot"


class EnvVariables(Enum):
    ATTACK_MODULES = "ATTACK_MODULES"
    BOOKMARKS = "BOOKMARKS"
    CONNECTORS = "CONNECTORS"
    CONNECTORS_ENDPOINTS = "CONNECTORS_ENDPOINTS"
    CONTEXT_STRATEGY = "CONTEXT_STRATEGY"
    COOKBOOKS = "COOKBOOKS"
    DATABASES = "DATABASES"
    DATABASES_MODULES = "DATABASES_MODULES"
    DATASETS = "DATASETS"
    IO_MODULES = "IO_MODULES"
    METRICS = "METRICS"
    PROMPT_TEMPLATES = "PROMPT_TEMPLATES"
    RECIPES = "RECIPES"
    RESULTS = "RESULTS"
    RESULTS_MODULES = "RESULTS_MODULES"
    RUNNERS = "RUNNERS"
    RUNNERS_MODULES = "RUNNERS_MODULES"


class EnvironmentVars:
    env_vars = {}

    ATTACK_MODULES = [
        env_vars.get(EnvVariables.ATTACK_MODULES.value),
        str(importlib.resources.files(__app_name__).joinpath("data/attack-modules")),
    ]
    BOOKMARKS = [
        env_vars.get(EnvVariables.BOOKMARKS.value),
        str(
            importlib.resources.files(__app_name__).joinpath(
                "data/generated-outputs/bookmarks"
            )
        ),
    ]
    CONNECTORS = [
        env_vars.get(EnvVariables.CONNECTORS.value),
        str(importlib.resources.files(__app_name__).joinpath("data/connectors")),
    ]
    CONNECTORS_ENDPOINTS = [
        env_vars.get(EnvVariables.CONNECTORS_ENDPOINTS.value),
        str(
            importlib.resources.files(__app_name__).joinpath(
                "data/connectors-endpoints"
            )
        ),
    ]
    CONTEXT_STRATEGY = [
        env_vars.get(EnvVariables.CONTEXT_STRATEGY.value),
        str(importlib.resources.files(__app_name__).joinpath("data/context-strategy")),
    ]
    COOKBOOKS = [
        env_vars.get(EnvVariables.COOKBOOKS.value),
        str(importlib.resources.files(__app_name__).joinpath("data/cookbooks")),
    ]
    DATABASES = [
        env_vars.get(EnvVariables.DATABASES.value),
        str(
            importlib.resources.files(__app_name__).joinpath(
                "data/generated-outputs/databases"
            )
        ),
    ]
    DATABASES_MODULES = [
        env_vars.get(EnvVariables.DATABASES_MODULES.value),
        str(importlib.resources.files(__app_name__).joinpath("data/databases-modules")),
    ]
    DATASETS = [
        env_vars.get(EnvVariables.DATASETS.value),
        str(importlib.resources.files(__app_name__).joinpath("data/datasets")),
    ]
    IO_MODULES = [
        env_vars.get(EnvVariables.IO_MODULES.value),
        str(importlib.resources.files(__app_name__).joinpath("data/io-modules")),
    ]
    METRICS = [
        env_vars.get(EnvVariables.METRICS.value),
        str(importlib.resources.files(__app_name__).joinpath("data/metrics")),
    ]
    PROMPT_TEMPLATES = [
        env_vars.get(EnvVariables.PROMPT_TEMPLATES.value),
        str(importlib.resources.files(__app_name__).joinpath("data/prompt-templates")),
    ]
    RECIPES = [
        env_vars.get(EnvVariables.RECIPES.value),
        str(importlib.resources.files(__app_name__).joinpath("data/recipes")),
    ]
    RESULTS = [
        env_vars.get(EnvVariables.RESULTS.value),
        str(
            importlib.resources.files(__app_name__).joinpath(
                "data/generated-outputs/results"
            )
        ),
    ]
    RESULTS_MODULES = [
        env_vars.get(EnvVariables.RESULTS_MODULES.value),
        str(importlib.resources.files(__app_name__).joinpath("data/results-modules")),
    ]
    RUNNERS = [
        env_vars.get(EnvVariables.RUNNERS.value),
        str(
            importlib.resources.files(__app_name__).joinpath(
                "data/generated-outputs/runners"
            )
        ),
    ]
    RUNNERS_MODULES = [
        env_vars.get(EnvVariables.RUNNERS_MODULES.value),
        str(importlib.resources.files(__app_name__).joinpath("data/runners-modules")),
    ]

    @staticmethod
    def load_env(env_dict: dict | None = None) -> None:
        """
        This method is used to load environment variables from a provided dictionary.
        If the dictionary is not provided, it defaults to an empty dictionary.

        The method updates the class attributes with the corresponding values from the dictionary.
        If a key from the class attributes is not found in the dictionary, it will not be updated and a
        message will be printed.

        Args:
            env_dict (dict | None): A dictionary containing the environment variables to be loaded.
                                    If None, an empty dictionary will be used.
        """
        if env_dict is None:
            env_dict = dict()

        keys = [e.value for e in EnvVariables]
        unset_keys = []
        for key in keys:
            if key in env_dict:
                given_path = Path(env_dict[key])
                if given_path.exists():
                    EnvironmentVars.__dict__[key][0] = str(given_path)
                else:
                    logger.warning(
                        f"Unable to set {key}. The provided path {given_path} does not exist. "
                        "The stock set will be used."
                    )
            else:
                unset_keys.append(key)
        if unset_keys:
            logger.warning(
                f"Unable to retrieve the following environment variables: {unset_keys}. The stock set will be used."
            )

    @staticmethod
    def get_file_path(file_type: str, file_name: str, ignore_existance: bool) -> str:
        """
        This method is used to get the file path for a given file type and file name.
        If the ignore existance flag is set to True, it returns the file path even if the file does not exist.

        Args:
            file_type (str): The type of the file (e.g., 'recipe', 'cookbook').
            file_name (str): The name of the file.
            ignore_existance (bool): A flag indicating whether to return the file path
                             even if the file does not exist.

        Returns:
            str: The file path of the file.
        """
        for directory in EnvironmentVars.get_file_directory(file_type):
            file_path = Path(directory) / file_name
            if ignore_existance:
                return str(file_path)
            else:
                if Path(file_path).exists():
                    return str(file_path)
        return ""

    @staticmethod
    def get_file_directory(file_type: str) -> list[str]:
        """
        This method retrieves the directory paths associated with a specified file type.

        Args:
            file_type (str): The type of file for which the directory paths are to be retrieved.

        Returns:
            list[str]: A list containing the directory paths associated with the given file type.
        """
        path_from_user, path_from_resource = getattr(EnvironmentVars, file_type)

        if path_from_user is not None:
            return [path_from_user, path_from_resource]
        else:
            return [path_from_resource]



import cmd2
from cmd2 import Statement
from rich.console import Console

from moonshot.integrations.cli.active_session_cfg import active_session
from moonshot.integrations.cli.benchmark.benchmark import BenchmarkCommandSet
from moonshot.integrations.cli.common.common import CommonCommandSet
from moonshot.integrations.cli.initialisation.initialisation import (
    InitialisationCommandSet,
)
from moonshot.integrations.cli.redteam.redteam import RedTeamCommandSet
from moonshot.integrations.cli.redteam.session import (
    manual_red_teaming,
    update_chat_display,
)

console = Console()


class CommandLineInterface(cmd2.Cmd):
    def __init__(self):
        super().__init__(terminators=[])
        self.prompt = "moonshot > "
        self.welcome()

    def welcome(self) -> None:
        """
        Display Project Moonshot logo
        """
        logo = "  _____           _           _     __  __                       _           _   \n"
        logo += " |  __ \\         (_)         | |   |  \\/  |                     | |         | |  \n"
        logo += " | |__) | __ ___  _  ___  ___| |_  | \\  / | ___   ___  _ __  ___| |__   ___ | |_ \n"
        logo += " |  ___/ '__/ _ \\| |/ _ \\/ __| __| | |\\/| |/ _ \\ / _ \\| '_ \\/ __| '_ \\ / _ \\| __|\n"
        logo += " | |   | | | (_) | |  __/ (__| |_  | |  | | (_) | (_) | | | \\__ \\ | | | (_) | |_ \n"
        logo += " |_|   |_|  \\___/| |\\___|\\___|\\__| |_|  |_|\\___/ \\___/|_| |_|___/_| |_|\\___/ \\__|\n"
        logo += "                _/ |                                                             \n"
        logo += "               |__/                                                              \n"
        logo += "\n"
        print(logo)

    def default(self, statement: Statement) -> None:
        # if there is an active session, anything entered other than recognised commands
        # will be considered prompts
        if active_session:
            user_prompt = statement.command + " " + statement
            user_prompt = user_prompt.strip()
            manual_red_teaming(user_prompt)
            # Update chat display with response
            update_chat_display()

    def postcmd(self, stop, line):
        if active_session:
            if active_session["context_strategy"]:
                cs_prompt = f"CS: ({active_session['context_strategy']},{active_session['cs_num_of_prev_prompts']})] > "
            else:
                cs_prompt = f"CS: {active_session['context_strategy']}]> "
            self.prompt = (
                f"moonshot ({active_session['session_id']}) "
                f"[PT: {active_session['prompt_template']}, "
                f"{cs_prompt}"
            )
        else:
            self.prompt = "moonshot > "
        return stop



import json
import os

from pydantic import validate_call

from moonshot.src.datasets.dataset import Dataset
from moonshot.src.datasets.dataset_arguments import DatasetArguments


# ------------------------------------------------------------------------------
# Datasets APIs
# ------------------------------------------------------------------------------
@validate_call
def api_delete_dataset(ds_id: str) -> bool:
    """
    Deletes a dataset identified by its unique dataset ID.

    Args:
        ds_id (str): The unique identifier for the dataset to be deleted.

    Returns:
        bool: True if the dataset was successfully deleted, False otherwise.

    Raises:
        Exception: If the deletion process encounters an error.
    """
    return Dataset.delete(ds_id)


def api_get_all_datasets() -> list[dict]:
    """
    This function retrieves all available datasets and returns them as a list of dictionaries. Each dictionary
    represents a dataset and contains its information.

    Returns:
        list[dict]: A list of dictionaries, each representing a dataset.
    """
    _, datasets = Dataset.get_available_items()
    return [dataset.to_dict() for dataset in datasets]


def api_get_all_datasets_name() -> list[str]:
    """
    This function retrieves all available dataset names and returns them as a list.

    Returns:
        list[str]: A list of dataset names.
    """
    datasets_name, _ = Dataset.get_available_items()
    return datasets_name


def api_download_dataset(
    name: str, description: str, reference: str, license: str, **kwargs
) -> str:
    """
    Downloads a dataset from Hugging Face and creates a new dataset with the provided details.

    This function takes the name, description, reference, and license for a new dataset as input, along with additional
    keyword arguments for downloading the dataset from Hugging Face. It then creates a new DatasetArguments object with
    these details and an empty id. The id is left empty because it will be generated from the name during the creation
    process. The function then calls the Dataset's create method to create the new dataset.

    Args:
        name (str): The name of the new dataset.
        description (str): A brief description of the new dataset.
        reference (str): A reference link for the new dataset.
        license (str): The license of the new dataset.
        kwargs: Additional keyword arguments for downloading the dataset from Hugging Face.

    Returns:
        str: The ID of the newly created dataset.
    """
    examples = Dataset.download_hf(**kwargs)
    ds_args = DatasetArguments(
        id="",
        name=name,
        description=description,
        reference=reference,
        license=license,
        examples=examples,
    )
    return Dataset.create(ds_args)


def api_convert_dataset(
    name: str, description: str, reference: str, license: str, file_path: str
) -> str:
    """
    Converts a CSV or JSON file to a dataset and creates a new dataset with the provided details.

    This function takes the name, description, reference, and license for a new dataset as input, along with the file
    path to a CSV file. It then creates a new DatasetArguments object with these details and an empty id. The id is left
    empty because it will be generated from the name during the creation process. The function then calls the Dataset's
    create method to create the new dataset.

    Args:
        name (str): The name of the new dataset.
        description (str): A brief description of the new dataset.
        reference (str): A reference link for the new dataset.
        license (str): The license of the new dataset.
        file_path (str): The file path to the CSV or JSONfile.

    Returns:
        str: The ID of the newly created dataset.
    """
    ds_args = None

    # Check if file is in a supported format
    if not (file_path.endswith(".json") or file_path.endswith(".csv")):
        raise ValueError("Unsupported file format. Please provide a JSON or CSV file.")

    # Check that file is not empty
    if os.path.getsize(file_path) == 0:
        raise ValueError("The uploaded file is empty.")

    # if file is already in json format
    if file_path.endswith(".json"):
        json_data = json.load(open(file_path))

        try:
            if "examples" in json_data and json_data["examples"]:
                ds_args = DatasetArguments(
                    id="",
                    name=json_data.get("name", name),
                    description=json_data.get("description", description),
                    reference=json_data.get("reference", reference),
                    license=json_data.get("license", license),
                    examples=iter(json_data["examples"]),
                )
            else:
                raise KeyError(
                    "examples is either empty or this key is not in the JSON file. "
                    "Please ensure that this field is present."
                )
        except Exception as e:
            raise e

    # if file is in csv format, convert data
    else:
        try:
            examples = Dataset.convert_data(file_path)
            ds_args = DatasetArguments(
                id="",
                name=name,
                description=description,
                reference=reference,
                license=license,
                examples=examples,
            )
        except Exception as e:
            raise e
    return Dataset.create(ds_args)



from dependency_injector.wiring import Provide, inject
from fastapi import APIRouter, Depends, HTTPException

from ..container import Container
from ..schemas.prompt_template_response_model import PromptTemplatesResponseModel
from ..services.prompt_template_service import PromptTemplateService
from ..services.utils.exceptions_handler import ServiceException

router = APIRouter(tags=["Prompt Template"])


@router.get("/api/v1/prompt-templates")
@inject
def get_all_prompt_templates(
    prompt_template_service: PromptTemplateService = Depends(
        Provide[Container.prompt_template_service]
    ),
) -> PromptTemplatesResponseModel:
    """
    Retrieve all prompt templates from the database.

    Args:
        prompt_template_service (PromptTemplateService): The service responsible for retrieving prompt templates.

    Returns:
        PromptTemplatesResponseModel: A model representing all prompt templates.

    Raises:
        HTTPException: An error with status code 404 if no prompt templates are found.
                       An error with status code 400 if there is a validation error.
                       An error with status code 500 for any other server-side error.
    """
    try:
        return prompt_template_service.get_prompt_templates()
    except ServiceException as e:
        if e.error_code == "FileNotFound":
            raise HTTPException(
                status_code=404, detail=f"Failed to retrieve prompt templates: {e.msg}"
            )
        elif e.error_code == "ValidationError":
            raise HTTPException(
                status_code=400, detail=f"Failed to retrieve prompt templates: {e.msg}"
            )
        else:
            raise HTTPException(
                status_code=500, detail=f"Failed to retrieve prompt templates: {e.msg}"
            )


@router.get("/api/v1/prompt-templates/name")
@inject
def get_all_prompt_templates_names(
    prompt_template_service: PromptTemplateService = Depends(
        Provide[Container.prompt_template_service]
    ),
) -> list[str]:
    """
    Retrieve the names of all prompt templates from the database.

    Args:
        prompt_template_service (PromptTemplateService): The service responsible for retrieving prompt template names.

    Returns:
        list[str]: A list of prompt template names.

    Raises:
        HTTPException: An error with status code 404 if no prompt template names are found.
                       An error with status code 400 if there is a validation error.
                       An error with status code 500 for any other server-side error.
    """
    try:
        return prompt_template_service.get_prompt_templates_name()
    except ServiceException as e:
        if e.error_code == "FileNotFound":
            raise HTTPException(
                status_code=404,
                detail=f"Failed to retrieve prompt template names: {e.msg}",
            )
        elif e.error_code == "ValidationError":
            raise HTTPException(
                status_code=400,
                detail=f"Failed to retrieve prompt template names: {e.msg}",
            )
        else:
            raise HTTPException(
                status_code=500,
                detail=f"Failed to retrieve prompt template names: {e.msg}",
            )


@router.delete("/api/v1/prompt-templates/{prompt_template_name}")
@inject
def delete_prompt_template(
    prompt_template_name: str,
    prompt_template_service: PromptTemplateService = Depends(
        Provide[Container.prompt_template_service]
    ),
) -> dict[str, bool]:
    """
    Delete a prompt template from the database by its name.

    Args:
        prompt_template_name (str): The name of the prompt template to delete.
        prompt_template_service (PromptTemplateService): The service responsible for deleting the prompt template.

    Returns:
        dict[str, bool]: A dictionary with a key 'success' indicating the result of the deletion operation.

    Raises:
        HTTPException: An error with status code 404 if the prompt template is not found.
                       An error with status code 400 if there is a validation error.
                       An error with status code 500 for any other server-side error.
    """
    try:
        prompt_template_service.delete_prompt_template(prompt_template_name)
        return {"success": True}
    except ServiceException as e:
        if e.error_code == "FileNotFound":
            raise HTTPException(
                status_code=404, detail=f"Failed to delete prompt template: {e.msg}"
            )
        elif e.error_code == "ValidationError":
            raise HTTPException(
                status_code=400, detail=f"Failed to delete prompt template: {e.msg}"
            )
        else:
            raise HTTPException(
                status_code=500, detail=f"Failed to delete prompt template: {e.msg}"
            )



from pydantic import BaseModel, RootModel

from ..schemas.session_response_model import ChatRecord


class PromptInfo(BaseModel):
    current_runner_id: str
    current_chats: dict[str, list[ChatRecord]]
    current_batch_size: int
    current_status: str


PromptResponseModel = RootModel[PromptInfo]



import time
from typing import Callable

from moonshot.src.runs.run_arguments import RunArguments
from moonshot.src.runs.run_status import RunStatus
from moonshot.src.storage.storage import Storage
from moonshot.src.utils.log import configure_logger

# Create a logger for this module
logger = configure_logger(__name__)


class RunProgress:
    sql_update_run_record = """
        UPDATE run_table SET runner_id=?,runner_type=?,runner_args=?,endpoints=?,results_file=?,start_time=?,end_time=?,
        duration=?,error_messages=?,raw_results=?,results=?,status=?
        WHERE run_id=?
    """

    def __init__(
        self, run_arguments: RunArguments, run_progress_callback_func: Callable | None
    ):
        # Information on the run and callback for progress updating
        self.run_arguments = run_arguments
        self.run_progress_callback_func = run_progress_callback_func

        # Information to be sent back through callback
        self.cookbook_index: int = -1
        self.cookbook_name: str = ""
        self.cookbook_total: int = -1
        self.recipe_index: int = -1
        self.recipe_name: str = ""
        self.recipe_total: int = -1
        self.progress: int = 0

    def notify_error(self, error_message: str) -> None:
        """
        Notifies about an error that occurred during the run process.

        This method logs the error message, appends it to the run arguments' error messages list,
        and updates the run progress status to indicate that the run is continuing with errors.

        Args:
            error_message (str): The error message to be logged and recorded.
        """
        # Update error message
        logger.error(error_message)
        self.run_arguments.error_messages.append(error_message)

        # Update progress status
        self.notify_progress(status=RunStatus.RUNNING_WITH_ERRORS)

    def notify_progress(self, **kwargs) -> None:
        """
        Updates the run progress information and the run status.

        This method is responsible for updating the run status, setting the end time, calculating the run duration,
        and persisting these changes to the database. Additionally, if a callback function for run progress has been
        set, this method will invoke it with the current state of run arguments.

        The method accepts arbitrary keyword arguments which are used to update specific attributes of the run_arguments
        object. It ensures that the run progress is accurately reflected in both the object's state and the
        corresponding database record.

        Args:
            **kwargs: Keyword arguments that correspond to the attributes of run_arguments which should be updated.

        Returns:
            None
        """
        # Update run arguments values
        if self.run_arguments.start_time > 0.0:
            self.run_arguments.end_time = time.time()
            self.run_arguments.duration = int(
                self.run_arguments.end_time - self.run_arguments.start_time
            )

        # Update run arguments values with provided key-value pairs
        for key, value in kwargs.items():
            # Update self values
            if hasattr(self, key):
                setattr(self, key, value)

            # Update self.run_arguments
            if hasattr(self.run_arguments, key):
                setattr(self.run_arguments, key, value)

        # Calculate percentage
        if self.cookbook_total > 0:
            if self.recipe_total > 0 and self.cookbook_index != self.cookbook_total:
                # Calculate percentage with cookbook and recipe defined
                per_recipe_percentage = (100 / self.cookbook_total) / self.recipe_total
                self.progress = int(
                    self.cookbook_index * (100 / self.cookbook_total)
                ) + int(self.recipe_index * per_recipe_percentage)
            else:
                # Calculate percentage with cookbook defined and no recipes defined
                # Or cookbook index and total is same.
                self.progress = int(self.cookbook_index * (100 / self.cookbook_total))
        elif self.recipe_total > 0:
            # There is no cookbook, calculate for recipes defined
            self.progress = int(self.recipe_index * (100 / self.recipe_total))
        else:
            # Initialization: set 0
            self.progress = 0

        # Update database record
        if self.run_arguments.database_instance:
            Storage.update_database_record(
                self.run_arguments.database_instance,
                self.run_arguments.to_tuple(),
                RunProgress.sql_update_run_record,
            )
        else:
            logger.warning(
                "[RunProgress] Failed to update run progress: database instance is not initialised."
            )

        # If a callback function is provided, call it with the updated run arguments
        if self.run_progress_callback_func:
            self.run_progress_callback_func(self.get_dict())

    def get_dict(self) -> dict:
        """
        Constructs and returns a dictionary with the current state of the benchmark execution.

        This method assembles a dictionary encapsulating the current progress of the benchmark execution.
        The resulting dictionary is composed of the following keys: execution identifier, name, type, current duration,
        status, cookbook index, cookbook name, cookbook total, recipe index, recipe name, recipe total,
        progress percentage, and a list of error messages.

        Returns:
            dict: A dictionary representing the current benchmark execution state, with keys for various progress
            metrics and details.
        """
        return {
            "current_runner_id": self.run_arguments.runner_id,
            "current_runner_type": self.run_arguments.runner_type.value,
            "current_duration": self.run_arguments.duration,
            "current_status": self.run_arguments.status.value,
            "current_cookbook_index": self.cookbook_index,
            "current_cookbook_name": self.cookbook_name,
            "current_cookbook_total": self.cookbook_total,
            "current_recipe_index": self.recipe_index,
            "current_recipe_name": self.recipe_name,
            "current_recipe_total": self.recipe_total,
            "current_progress": self.progress,
            "current_error_messages": list(set(self.run_arguments.error_messages)),
        }



from typing import Callable

from moonshot.src.runners.runner import Runner

from .... import api as moonshot_api
from ..services.base_service import BaseService
from ..services.utils.exceptions_handler import exception_handler
from ..types.types import TestRunProgress


class RunnerService(BaseService):
    @exception_handler
    def get_all_runner(self) -> list[dict]:
        """
        Retrieves all runners.

        Returns:
            A list of dictionaries, where each dictionary contains the details of a runner.
        """
        runners = moonshot_api.api_get_all_runner()
        return runners

    @exception_handler
    def get_all_runner_name(self) -> list[str]:
        """
        Retrieves the names of all runners.

        Returns:
            A list of strings, where each string is the name of a runner.
        """
        runners = moonshot_api.api_get_all_runner_name()
        return runners

    @exception_handler
    def get_runner_by_id(self, runner_id: str) -> dict | None:
        """
        Retrieves a runner by its ID.

        Args:
            runner_id: The unique identifier of the runner.

        Returns:
            A dictionary containing the runner's details if found, otherwise None.
        """
        runner = moonshot_api.api_read_runner(runner_id)
        return runner

    @exception_handler
    def delete_run(self, runner_id: str) -> None:
        """
        Deletes a runner by its ID.

        Args:
            runner_id: The unique identifier of the runner to be deleted.
        """
        moonshot_api.api_delete_runner(runner_id)

    @exception_handler
    def create_runner(
        self,
        runner_name: str,
        endpoints: list[str],
        description: str,
        progress_callback_func: Callable[[TestRunProgress], None] = None,
    ) -> Runner:
        """
        Creates a new runner with the specified details.

        Args:
            runner_name: The name of the runner.
            endpoints: A list of endpoints the runner will interact with.
            description: A brief description of the runner.
            progress_callback_func: An optional callback function for progress updates.

        Returns:
            An instance of Runner.
        """
        return moonshot_api.api_create_runner(
            name=runner_name,
            endpoints=endpoints,
            description=description,
            progress_callback_func=progress_callback_func,
        )

    @exception_handler
    def load_runner(
        self,
        runner_id: str,
        progress_callback_func: Callable[[TestRunProgress], None] = None,
    ) -> Runner:
        """
        Loads a runner by its ID.

        Args:
            runner_id: The unique identifier of the runner to be loaded.
            progress_callback_func: An optional callback function for progress updates.

        Returns:
            An instance of Runner.
        """
        return moonshot_api.api_load_runner(
            runner_id, progress_callback_func=progress_callback_func
        )

    @exception_handler
    def get_run_details_by_runner(self, runner_id: str, run_id: int):
        """
        Retrieves the details of a specific run by the runner ID and run ID.

        Args:
            runner_id: The unique identifier of the runner.
            run_id: The unique identifier of the run.

        Returns:
            A dictionary containing the details of the run if found, otherwise an empty dictionary.
        """
        runs = moonshot_api.api_get_all_run(runner_id)
        retn = {}
        run_id_int = int(run_id)
        for run in runs:
            run_id_from_run = int(run.get("run_id"))
            if run_id_from_run == run_id_int:
                retn = {
                    "run_id": run.get("run_id"),
                    "runner_id": run.get("runner_id"),
                    "runner_args": run.get("runner_args"),
                    "endpoints": run.get("endpoints"),
                    "start_time": run.get("start_time"),
                }
        return retn

    @exception_handler
    def get_runs_id_in_runner(self, runner_id) -> list[int]:
        """
        Retrieves a list of run IDs for a given runner.

        Args:
            runner_id: The unique identifier of the runner.

        Returns:
            A list of integers, where each integer is the ID of a run associated with the runner.
        """
        runs = moonshot_api.api_get_all_run(runner_id)
        retn = []
        for run in runs:
            retn.append(run.get("run_id"))
        return retn



from ast import literal_eval

import cmd2
from rich.console import Console
from rich.table import Table

from moonshot.api import api_get_all_run
from moonshot.integrations.cli.cli_errors import (
    ERROR_BENCHMARK_LIST_RUNS_FIND_VALIDATION,
    ERROR_BENCHMARK_LIST_RUNS_PAGINATION_VALIDATION,
    ERROR_BENCHMARK_LIST_RUNS_PAGINATION_VALIDATION_1,
    ERROR_BENCHMARK_VIEW_RUN_RUNNER_ID_VALIDATION,
)
from moonshot.integrations.cli.common.display_helper import (
    display_view_list_format,
    display_view_str_format,
)
from moonshot.integrations.cli.utils.process_data import filter_data

console = Console()


# ------------------------------------------------------------------------------
# CLI Functions
# ------------------------------------------------------------------------------
def list_runs(args) -> list | None:
    """
    List all runs.

    This function retrieves all available runs by calling the api_get_all_run function from the
    moonshot.api module. It then filters the runs based on the provided keyword and pagination arguments.
    If there are no runs, it prints a message indicating that no runs were found.

    Args:
        args (argparse.Namespace): The arguments provided to the command line interface.
            find (str): Optional field to find run(s) with a keyword.
            pagination (str): Optional field to paginate runs.

    Returns:
        list | None: A list of runs or None if there are no runs.
    """

    try:
        if args.find is not None:
            if not isinstance(args.find, str) or not args.find:
                raise TypeError(ERROR_BENCHMARK_LIST_RUNS_FIND_VALIDATION)

        if args.pagination is not None:
            if not isinstance(args.pagination, str) or not args.pagination:
                raise TypeError(ERROR_BENCHMARK_LIST_RUNS_PAGINATION_VALIDATION)
            try:
                pagination = literal_eval(args.pagination)
                if not (
                    isinstance(pagination, tuple)
                    and len(pagination) == 2
                    and all(isinstance(i, int) for i in pagination)
                ):
                    raise ValueError(ERROR_BENCHMARK_LIST_RUNS_PAGINATION_VALIDATION_1)
            except (ValueError, SyntaxError):
                raise ValueError(ERROR_BENCHMARK_LIST_RUNS_PAGINATION_VALIDATION_1)
        else:
            pagination = ()

        runner_run_info = api_get_all_run()
        keyword = args.find.lower() if args.find else ""

        if runner_run_info:
            filtered_runs_list = filter_data(runner_run_info, keyword, pagination)
            if filtered_runs_list:
                _display_runs(filtered_runs_list)
                return filtered_runs_list

        console.print("[red]There are no runs found.[/red]")
        return None

    except Exception as e:
        print(f"[list_runs]: {str(e)}")
        return None


def view_run(args) -> None:
    """
    View the details of a specific run.

    This function retrieves and displays information about a specific run associated with a runner. It uses the runner
    identifier provided in the arguments to fetch the data and then calls the _display_runs function to present it in a
    user-friendly format.

    Args:
        args (argparse.Namespace): The arguments provided to the command line interface.
            runner_id (str): The identifier of the runner whose runs are to be viewed.

    Returns:
        None
    """
    try:
        if (
            not isinstance(args.runner_id, str)
            or not args.runner_id
            or args.runner_id is None
        ):
            raise TypeError(ERROR_BENCHMARK_VIEW_RUN_RUNNER_ID_VALIDATION)

        runner_run_info = api_get_all_run(args.runner_id)
        _display_runs(runner_run_info)
    except Exception as e:
        print(f"[view_run]: {str(e)}")


# ------------------------------------------------------------------------------
# Helper functions: Display on cli
# ------------------------------------------------------------------------------
def _display_runs(runs_list: list):
    """
    Display a list of runs in a table format.

    This function takes a list of run information and displays it in a table format using the rich library's
    Table object.

    Each run's details are formatted and added as a row in the table.
    If there are no runs to display, a message is printed to indicate that no results were found.

    Args:
        runs_list (list): A list of dictionaries, where each dictionary contains details of a run.

    Returns:
        None
    """
    table = Table(
        title="List of Runs", show_lines=True, expand=True, header_style="bold"
    )
    table.add_column("No.", width=2)
    table.add_column("Run", justify="left", width=78)
    table.add_column("Contains", justify="left", width=20, overflow="fold")
    for idx, run in enumerate(runs_list, 1):
        (
            run_id,
            runner_id,
            runner_type,
            runner_args,
            endpoints,
            results_file,
            start_time,
            end_time,
            duration,
            error_messages,
            raw_results,
            results,
            status,
            *other_args,
        ) = run.values()

        duration_info = f"[blue]Period:[/blue] {start_time} - {end_time} ({duration}s)"
        run_id = display_view_str_format("Run ID", run_id)
        runner_id = display_view_str_format("Runner ID", runner_id)
        runner_type = display_view_str_format("Runner Type", runner_type)
        runner_args = display_view_str_format("Runner Args", runner_args)
        status_info = display_view_str_format("Status", status)
        results_info = display_view_str_format("Results File", results_file)
        endpoints_info = display_view_list_format("Endpoints", endpoints)
        error_messages_info = display_view_list_format("Error Messages", error_messages)

        has_raw_results = bool(raw_results)
        has_results = bool(results)
        idx = run.get("idx", idx)

        result_info = (
            f"[red]{runner_id}[/red]\n\n{run_id}\n\n{duration_info}\n\n{status_info}"
        )
        contains_info = (
            f"{results_info}\n\n{error_messages_info}\n\n{endpoints_info}\n\n"
            f"[blue]Has Raw Results: {has_raw_results}[/blue]\n\n"
            f"[blue]Has Results: {has_results}[/blue]"
        )

        table.add_section()
        table.add_row(str(idx), result_info, contains_info)
    console.print(table)


# ------------------------------------------------------------------------------
# Cmd2 Arguments Parsers
# ------------------------------------------------------------------------------
# View run arguments
view_run_args = cmd2.Cmd2ArgumentParser(
    description="View a runner's runs.",
    epilog="Example:\n view_run my-new-cookbook-runner",
)
view_run_args.add_argument("runner_id", type=str, help="Name of the runner")


# List run arguments
list_runs_args = cmd2.Cmd2ArgumentParser(
    description="List all runs.",
    epilog='Example:\n list_runs -f "my-run"',
)

list_runs_args.add_argument(
    "-f",
    "--find",
    type=str,
    help="Optional field to find run(s) with keyword",
    nargs="?",
)

list_runs_args.add_argument(
    "-p",
    "--pagination",
    type=str,
    help="Optional tuple to paginate run(s). E.g. (2,10) returns 2nd page with 10 items in each page.",
    nargs="?",
)



from pydantic import conlist, validate_call

from moonshot.src.results.result import Result


# ------------------------------------------------------------------------------
# Result APIs
# ------------------------------------------------------------------------------
@validate_call
def api_read_result(res_id: str) -> dict:
    """
    Reads a result and returns its information.

    This function takes a result ID as input, reads the corresponding result from the storage manager,
    and returns a dictionary containing the result's information.

    Args:
        res_id (str): The ID of the result.

    Returns:
        dict: A dictionary containing the result's information.
    """
    return Result.read(res_id)


@validate_call
def api_read_results(res_ids: conlist(str, min_length=1)) -> list[dict]:
    """
    Reads multiple results and returns their information.

    This function takes a list of result IDs as input, reads the corresponding results from the storage manager,
    and returns a list of dictionaries, each containing a result's information.

    Args:
        res_ids (conlist(str, min_length=1)): A list of result IDs.

    Returns:
        list[dict]: A list of dictionaries, each containing a result's information.
    """
    return [Result.read(res_id) for res_id in res_ids]


@validate_call
def api_delete_result(res_id: str) -> bool:
    """
    Deletes a result by its identifier.

    This function takes a result ID as input and calls the delete method from the Result class
    to remove the specified result from storage.

    Args:
        res_id (str): The unique identifier of the result to be deleted.

    Returns:
        bool: True if the result was successfully deleted.

    Raises:
        Exception: If the deletion process encounters an error.
    """
    return Result.delete(res_id)


def api_get_all_result() -> list[dict]:
    """
    This function retrieves all available results and returns them as a list of dictionaries.

    This function calls the get_available_items method from the Result class to retrieve all available results.
    It then returns a list of dictionaries, each containing the details of a result.

    Returns:
        list[dict]: A list of dictionaries, each representing a result's details.
    """
    _, results = Result.get_available_items()
    return [result for result in results]


def api_get_all_result_name() -> list[str]:
    """
    This function retrieves all available result names and returns them as a list.

    This function calls the get_available_items method from the Result class to retrieve all available results.
    It then extracts the names of each result and returns a list of these names.

    Returns:
        list[str]: A list of result names.
    """
    results_name, _ = Result.get_available_items()
    return results_name



from __future__ import annotations

from pathlib import Path

from pydantic import validate_call
from slugify import slugify

from moonshot.src.configs.env_variables import EnvVariables
from moonshot.src.cookbooks.cookbook_arguments import CookbookArguments
from moonshot.src.recipes.recipe import Recipe
from moonshot.src.storage.storage import Storage
from moonshot.src.utils.log import configure_logger

# Create a logger for this module
logger = configure_logger(__name__)


class Cookbook:
    def __init__(self, cb_args: CookbookArguments) -> None:
        self.id = cb_args.id
        self.name = cb_args.name
        self.description = cb_args.description
        self.tags = cb_args.tags
        self.categories = cb_args.categories
        self.recipes = cb_args.recipes

    @classmethod
    def load(cls, cb_id: str) -> Cookbook:
        """
        This method loads a cookbook from a JSON file.

        It uses the cookbook ID to construct the file path for the JSON file in the designated cookbook directory.
        The method then reads the JSON file and returns the cookbook information as a Cookbook instance.

        Args:
            cb_id (str): The unique identifier of the cookbook.

        Returns:
            Cookbook: An instance of the Cookbook class populated with the loaded cookbook information.
        """
        return cls(Cookbook.read(cb_id))

    @staticmethod
    def create(cb_args: CookbookArguments) -> str:
        """
        This method is responsible for creating a new cookbook and storing its details in a JSON file.

        The function accepts `cb_args` parameter which contains the necessary details for creating a new cookbook.
        It generates a unique ID for the cookbook by slugifying the cookbook name. After that, it constructs a
        dictionary with the cookbook's details and writes this information to a JSON file. The JSON file is named after
        the cookbook ID and is stored in the directory specified by `EnvironmentVars.COOKBOOKS`.

        If the operation encounters any error, an exception is raised and the error message is printed.

        Args:
            cb_args (CookbookArguments): An object that holds the necessary details for creating a new cookbook.

        Returns:
            str: The unique ID of the newly created cookbook.

        Raises:
            RuntimeError: If any of the recipes specified in the cookbook does not exist.
            Exception: If there is an error during the file writing process or any other operation within the method.
        """
        try:
            cb_id = slugify(cb_args.name, lowercase=True)
            cb_info = {
                "name": cb_args.name,
                "description": cb_args.description,
                "tags": Cookbook.get_tags_in_recipes(cb_args.recipes),
                "categories": Cookbook.get_categories_in_recipes(cb_args.recipes),
                "recipes": cb_args.recipes,
            }

            # check if the cookbook exists
            if Storage.is_object_exists(EnvVariables.COOKBOOKS.name, cb_id, "json"):
                raise RuntimeError(f"Cookbook with ID '{cb_id}' already exists.")

            # check if recipes in list exist before creating cookbook
            for recipe in cb_args.recipes:
                if not Storage.is_object_exists(
                    EnvVariables.RECIPES.name, recipe, "json"
                ):
                    raise RuntimeError(f"{recipe} recipe does not exist.")

            # Write as json output
            Storage.create_object(EnvVariables.COOKBOOKS.name, cb_id, cb_info, "json")
            return cb_id

        except Exception as e:
            logger.error(f"Failed to create cookbook: {str(e)}")
            raise e

    @staticmethod
    @validate_call
    def read(cb_id: str) -> CookbookArguments:
        """
        Fetches and returns the details of a specified cookbook by its ID.

        This method takes a cookbook ID, searches for its corresponding JSON file in the directory set by
        `EnvironmentVars.COOKBOOKS`, and constructs a CookbookArguments object with the cookbook's details.

        If the process encounters any issues, such as the file not existing or being inaccessible, it logs the error
        and raises an exception.

        Args:
            cb_id (str): The unique identifier of the cookbook to fetch.

        Returns:
            CookbookArguments: An instance filled with the cookbook's details.

        Raises:
            RuntimeError: If the cookbook ID is empty or the specified cookbook does not exist.
            Exception: For any issues encountered during the file reading or data parsing process.
        """
        try:
            if not cb_id:
                raise RuntimeError("Cookbook ID is empty.")

            cookbook_details = Cookbook._read_cookbook(cb_id)
            if not cookbook_details:
                raise RuntimeError(f"Cookbook with ID '{cb_id}' does not exist.")

            return CookbookArguments(**cookbook_details)

        except Exception as e:
            logger.error(f"Failed to read cookbook: {str(e)}")
            raise

    @staticmethod
    def _read_cookbook(cb_id: str) -> dict:
        """
        Retrieves the cookbook's information from a JSON file.

        This internal method is designed to fetch the details of a specific cookbook by its ID. It searches for the
        corresponding JSON file within the directory specified by `EnvVariables.COOKBOOKS`. The method returns a
        dictionary containing the cookbook's information.

        Args:
            cb_id (str): The unique identifier of the cookbook whose information is being retrieved.

        Returns:
            dict: A dictionary with the cookbook's information.
        """
        cookbook_info = {"id": cb_id}
        cookbook_info.update(
            Storage.read_object(EnvVariables.COOKBOOKS.name, cb_id, "json")
        )
        return cookbook_info

    @staticmethod
    def update(cb_args: CookbookArguments) -> bool:
        """
        Updates the details of an existing cookbook.

        This method accepts a CookbookArguments object, converts it to a dictionary, and writes the updated
        information to the corresponding JSON file in the directory defined by `EnvVariables.COOKBOOKS`.

        Args:
            cb_args (CookbookArguments): An object containing the updated details of the cookbook.

        Returns:
            bool: True if the update was successful.

        Raises:
            Exception: If there's an error during the update process.
        """
        try:
            # check if recipes in list exist before creating cookbook
            for recipe in cb_args.recipes:
                if not Storage.is_object_exists(
                    EnvVariables.RECIPES.name, recipe, "json"
                ):
                    raise RuntimeError(f"{recipe} recipe does not exist.")

            # Serialize the CookbookArguments object to a dictionary and remove derived properties
            cb_info = cb_args.to_dict()
            cb_info.pop("id", None)  # The 'id' is derived and should not be written

            # Write the updated cookbook information to the storage
            Storage.create_object(
                EnvVariables.COOKBOOKS.name, cb_args.id, cb_info, "json"
            )
            return True

        except Exception as e:
            logger.error(f"Failed to update cookbook: {str(e)}")
            raise e

    @staticmethod
    @validate_call
    def delete(cb_id: str) -> bool:
        """
        Deletes a cookbook identified by its ID.

        This method removes the cookbook's JSON file from the storage, using the `Storage.delete_object` method.
        The `EnvVariables.COOKBOOKS` environment variable specifies the directory where the cookbook files are stored.

        Args:
            cb_id (str): The unique identifier of the cookbook to be deleted.

        Returns:
            bool: True if the deletion was successful.

        Raises:
            Exception: If there's an error during the deletion process.
        """
        try:
            Storage.delete_object(EnvVariables.COOKBOOKS.name, cb_id, "json")
            return True

        except Exception as e:
            logger.error(f"Failed to delete cookbook: {str(e)}")
            raise e

    @staticmethod
    def get_available_items() -> tuple[list[str], list[CookbookArguments]]:
        """
        Retrieves and returns all available cookbooks.

        This method scans the directory specified by `EnvironmentVars.COOKBOOKS` and identifies all stored cookbook
        files. It excludes any files that contain "__" in their names. For each valid cookbook file, the method reads
        the file content and constructs a CookbookArguments object encapsulating the cookbook's details.
        Both the CookbookArguments object and the cookbook ID are then appended to their respective lists.

        Returns:
            tuple[list[str], list[CookbookArguments]]: A tuple where the first element is a list of cookbook IDs and
            the second element is a list of CookbookArguments objects representing the details of each cookbook.

        Raises:
            Exception: If an error occurs during the file reading process or any other operation within the method.
        """
        try:
            retn_cbs = []
            retn_cbs_ids = []

            cbs = Storage.get_objects(EnvVariables.COOKBOOKS.name, "json")
            for cb in cbs:
                if "__" in cb:
                    continue

                cb_info = CookbookArguments(**Cookbook._read_cookbook(Path(cb).stem))
                retn_cbs.append(cb_info)
                retn_cbs_ids.append(cb_info.id)

            return retn_cbs_ids, retn_cbs

        except Exception as e:
            logger.error(f"Failed to get available cookbooks: {str(e)}")
            raise e

    @staticmethod
    def get_categories_in_recipes(recipes: list[str]) -> list[str]:
        return list(
            {
                category
                for recipe_id in recipes
                for category in Recipe.read(recipe_id).categories
            }
        )

    @staticmethod
    def get_tags_in_recipes(recipes: list[str]) -> list[str]:
        return list(
            {tag for recipe_id in recipes for tag in Recipe.read(recipe_id).tags}
        )



from pydantic import BaseModel, ConfigDict, Field


class BenchmarkRunnerDTO(BaseModel):
    model_config = ConfigDict(from_attributes=True)
    run_name: str
    description: str
    endpoints: list[str]
    inputs: list[str]
    prompt_selection_percentage: int = Field(..., ge=1, le=100)
    random_seed: int
    system_prompt: str
    runner_processing_module: str



from .... import api as moonshot_api
from ..services.base_service import BaseService
from ..services.utils.exceptions_handler import exception_handler
from ..types.types import BenchmarkResult


class BenchmarkResultService(BaseService):
    @exception_handler
    def get_all_results(self) -> list[BenchmarkResult]:
        results = moonshot_api.api_get_all_result()
        return [BenchmarkResult(**result) for result in results]

    @exception_handler
    def get_all_result_name(self) -> list[str] | None:
        results = moonshot_api.api_get_all_result_name()
        return results

    @exception_handler
    def get_result_by_id(self, result_id: str) -> BenchmarkResult:
        result = moonshot_api.api_read_result(result_id)
        return BenchmarkResult(**result)

    @exception_handler
    def delete_result(self, result_id: str) -> None:
        moonshot_api.api_delete_result(result_id)



import os
import shutil

# Global variable to store the path to the temporary folder
TEMP_FOLDER_PATH = None


def create_temp_dir(temp_file_path: str):
    """
    Create a temporary directory at the specified path.

    This function sets a global variable with the path to the temporary directory
    and creates the directory if it does not already exist.

    Args:
        temp_file_path (str): The file path where the temporary directory will be created.
    """
    global TEMP_FOLDER_PATH
    TEMP_FOLDER_PATH = temp_file_path
    if not os.path.exists(TEMP_FOLDER_PATH):
        os.makedirs(TEMP_FOLDER_PATH)


def copy_file(file_path: str) -> str:
    """
    Copy a file to the temporary directory.

    This function copies a file from the given file path to the temporary directory
    set by the create_temp_dir function. It raises an error if the temporary directory
    path is not set before calling this function.

    Args:
        file_path (str): The path of the file to be copied.

    Returns:
        str: The path to the copied file in the temporary directory.

    Raises:
        ValueError: If the temporary folder path is not set.
    """
    if TEMP_FOLDER_PATH is None:
        raise ValueError(
            "Temporary folder path is not set. Call create_temp_dir first."
        )

    # Define the destination path for the copied file
    dest_path = os.path.join(TEMP_FOLDER_PATH, os.path.basename(file_path))

    # Copy the file to the destination
    shutil.copy(file_path, dest_path)

    return dest_path



import asyncio
import uuid
import traceback
from typing import Any

from dependency_injector.wiring import inject

from moonshot.src.runners.runner import Runner

from ..schemas.benchmark_runner_dto import BenchmarkRunnerDTO
from ..services.base_service import BaseService
from ..services.benchmark_test_state import BenchmarkTestState
from ..services.runner_service import RunnerService
from ..status_updater.interface.benchmark_progress_callback import (
    InterfaceBenchmarkProgressCallback,
)
from ..types.types import BenchmarkCollectionType, TestRunProgress


class BenchmarkTestManager(BaseService):
    @inject
    def __init__(
        self,
        benchmark_test_state: BenchmarkTestState,
        progress_status_updater: InterfaceBenchmarkProgressCallback,
        runner_service: RunnerService,
    ) -> None:
        self.benchmark_test_state = benchmark_test_state
        self.progress_status_updater = progress_status_updater
        self.runner_service = runner_service
        super().__init__()

    def generate_unique_task_id(self) -> str:
        unique_id = str(uuid.uuid4())
        return f"task_{unique_id}"

    def add_task(
        self, executor_id: str, task: asyncio.Task[Any], moonshot_runner: Runner
    ) -> None:
        self.benchmark_test_state.add_task(executor_id, task, moonshot_runner)

    def remove_task(self, executor_id: str) -> None:
        self.benchmark_test_state.remove_task(executor_id)

    async def cancel_task(self, executor_id: str) -> None:
        await self.benchmark_test_state.cancel_task(executor_id)

    def update_progress_status(self, updates: TestRunProgress):
        self.benchmark_test_state.update_progress_status(updates)

    def on_task_completed(self, task: asyncio.Task[Any]) -> None:
        self.logger.debug(f"Task {task.get_name()} has completed")

    async def run_test(
        self,
        benchmark_input_data: BenchmarkRunnerDTO,
        benchmark_type: BenchmarkCollectionType,
        moonshot_runner: Runner,
    ) -> None:
        try:
            if benchmark_type == BenchmarkCollectionType.COOKBOOK:
                async_run = moonshot_runner.run_cookbooks(
                    cookbooks=benchmark_input_data.inputs,
                    prompt_selection_percentage=benchmark_input_data.prompt_selection_percentage,
                    random_seed=benchmark_input_data.random_seed,
                    system_prompt=benchmark_input_data.system_prompt,
                )
            else:
                async_run = moonshot_runner.run_recipes(
                    recipes=benchmark_input_data.inputs,
                    prompt_selection_percentage=benchmark_input_data.prompt_selection_percentage,
                    random_seed=benchmark_input_data.random_seed,
                    system_prompt=benchmark_input_data.system_prompt,
                )
        except Exception as e:
            self.logger.error(f"Failed to execute benchmark - {e}")
            raise Exception(f"Unexpected error in core library - {e}")

        await async_run
        async_run.close()

    async def schedule_test_task(
        self, input_data: BenchmarkRunnerDTO, benchmark_type: BenchmarkCollectionType
    ) -> str:
        task_id = self.generate_unique_task_id()
        try:
            runner = self.runner_service.create_runner(
                input_data.run_name,
                input_data.endpoints,
                input_data.description,
                self.progress_status_updater.on_progress_update,
            )
        except ValueError as e:
            self.logger.error("Failed to create runner - Error: %s, Stacktrace: %s", e, traceback.print_exc())
            raise ValueError("Failed to create runner due to invalid input.")
        except Exception as e:
            self.logger.error("Failed to create runner - Error: %s, Stacktrace: %s", e, traceback.print_exc())
            raise Exception(f"Unexpected error in core library - {e}")

        benchmark_coroutine = self.run_test(input_data, benchmark_type, runner)

        task = asyncio.create_task(benchmark_coroutine, name=task_id)

        def on_executor_completion(task: asyncio.Task[Any]):
            if task.exception():
                self.logger.error(
                    f"Executor {task.get_name()} has failed - {task.exception()}"
                )
            else:
                self.logger.debug(f"Executor {task.get_name()} has completed")

        task.add_done_callback(on_executor_completion)

        self.add_task(runner.id, task, runner)
        return runner.id



from __future__ import annotations

import asyncio
import time
from ast import literal_eval
from datetime import datetime
from typing import Any, Callable

from slugify import slugify

from moonshot.src.configs.env_variables import EnvVariables
from moonshot.src.redteaming.session.chat import Chat
from moonshot.src.redteaming.session.red_teaming_progress import RedTeamingProgress
from moonshot.src.redteaming.session.red_teaming_type import RedTeamingType
from moonshot.src.runners.runner_type import RunnerType
from moonshot.src.runs.run_status import RunStatus
from moonshot.src.storage.db_interface import DBInterface
from moonshot.src.storage.storage import Storage
from moonshot.src.utils.import_modules import get_instance
from moonshot.src.utils.log import configure_logger

# Create a logger for this module
logger = configure_logger(__name__)


class SessionMetadata:
    # TODO: convert this into a pydantic model
    def __init__(
        self,
        session_id: str,
        endpoints: list[str],
        created_epoch: float,
        created_datetime: str,
        prompt_template: str,
        context_strategy: str,
        cs_num_of_prev_prompts: int,
        attack_module: str,
        metric: str,
        system_prompt: str,
    ):
        self.session_id = self.check_type(session_id, str)
        self.endpoints = self.check_type(endpoints, list)
        self.created_epoch = self.check_type(created_epoch, float)
        self.created_datetime = self.check_type(created_datetime, str)
        self.prompt_template = self.check_type(prompt_template, str)
        self.context_strategy = self.check_type(context_strategy, str)
        self.cs_num_of_prev_prompts = self.check_type(cs_num_of_prev_prompts, int)
        self.attack_module = self.check_type(attack_module, str)
        self.metric = self.check_type(metric, str)
        self.system_prompt = self.check_type(system_prompt, str)

    def to_dict(self) -> dict:
        """
        Converts the SessionMetadata instance into a dictionary.

        Returns:
            dict: A dictionary representation of the SessionMetadata instance.
        """
        return {
            "session_id": self.session_id,
            "endpoints": self.endpoints,
            "created_epoch": str(self.created_epoch),
            "created_datetime": self.created_datetime,
            "prompt_template": self.prompt_template,
            "context_strategy": self.context_strategy,
            "cs_num_of_prev_prompts": self.cs_num_of_prev_prompts,
            "attack_module": self.attack_module,
            "metric": self.metric,
            "system_prompt": self.system_prompt,
        }

    def to_tuple(self) -> tuple:
        """
        Converts the SessionMetadata instance into a tuple.

        Returns:
            tuple: A tuple representation of the SessionMetadata instance.
        """
        return (
            self.session_id,
            str(self.endpoints),
            self.created_epoch,
            self.created_datetime,
            self.prompt_template,
            self.context_strategy,
            self.cs_num_of_prev_prompts,
            self.attack_module,
            self.metric,
            self.system_prompt,
        )

    @classmethod
    def from_tuple(cls, data_tuple: tuple) -> SessionMetadata:
        """
        Creates a SessionMetadata instance from a tuple using the class method.

        Args:
            data_tuple (tuple): A tuple containing session_id, endpoints, created_epoch, and created_datetime.

        Returns:
            SessionMetadata: An instance of SessionMetadata.
        """
        (
            runner_id,
            endpoints,
            created_epoch,
            created_datetime,
            prompt_template,
            context_strategy,
            cs_num_of_prev_prompts,
            attack_module,
            metric,
            system_prompt,
        ) = data_tuple

        return cls(
            runner_id,
            literal_eval(endpoints),
            created_epoch,
            created_datetime,
            prompt_template,
            context_strategy,
            cs_num_of_prev_prompts,
            attack_module,
            metric,
            system_prompt,
        )

    def check_type(self, checked_attribute: Any, expected_type: type) -> Any:
        """
        Checks if the type of the given attribute matches the expected type.

        Args:
            checked_attribute (Any): The attribute to be checked.
            expected_type (type): The expected type of the attribute.

        Returns:
            Any: The checked attribute if its type matches the expected type.

        Raises:
            TypeError: If the type of the checked attribute does not match the expected type.
        """
        if not isinstance(checked_attribute, expected_type):
            raise TypeError(
                f"Expected type for {checked_attribute} is {expected_type}, but got {type(checked_attribute)}"
            )
        return checked_attribute


class Session:
    DEFAULT_CONTEXT_STRATEGY_PROMPT = 5
    sql_create_session_metadata_table = """
            CREATE TABLE IF NOT EXISTS session_metadata_table (
            session_id text PRIMARY KEY NOT NULL,
            endpoints text NOT NULL,
            created_epoch INTEGER NOT NULL,
            created_datetime text NOT NULL,
            prompt_template text,
            context_strategy text,
            cs_num_of_prev_prompts int,
            attack_module text,
            metric text,
            system_prompt text
            );
    """

    sql_create_chat_history_table = """
        CREATE TABLE IF NOT EXISTS {} (
        id INTEGER PRIMARY KEY AUTOINCREMENT,
        connection_id text NOT NULL,
        context_strategy text,
        prompt_template text,
        attack_module text,
        metric text,
        prompt text NOT NULL,
        prepared_prompt text NOT NULL,
        system_prompt text,
        predicted_result text NOT NULL,
        duration text NOT NULL,
        prompt_time text NOT NULL
        );
    """

    sql_create_session_metadata_record = """
        INSERT INTO session_metadata_table (
        session_id,endpoints,created_epoch,created_datetime,prompt_template,context_strategy,cs_num_of_prev_prompts,
        attack_module, metric, system_prompt) VALUES(?,?,?,?,?,?,?,?,?,?)
    """

    sql_read_session_metadata = """
        SELECT * from session_metadata_table
    """

    sql_update_session_metadata_field = """
        UPDATE session_metadata_table SET {}=? WHERE session_id=?
    """

    sql_drop_table = """
        DROP TABLE IF EXISTS {}
    """

    def __init__(
        self,
        runner_id: str,
        runner_type: RunnerType,
        runner_args: dict,
        database_instance: Any | None,
        endpoints: list[str],
        results_file_path: str,
        progress_callback_func: Callable | None = None,
    ):
        """
        Initializes a new session with the given parameters, creates session metadata,
        and sets up the database tables for session metadata and chat history.

        Args:
            runner_id (str): The unique identifier for the runner.
            runner_type (RunnerType): The type of runner being used.
            runner_args (dict): A dictionary of arguments specific to the runner.
            database_instance (Any | None): The database instance to connect to, or None if not available.
            endpoints (list[str]): A list of endpoint identifiers.
            results_file_path (str): The file path where results should be stored.
            progress_callback_func (Callable | None): An optional callback function for progress updates.
        """
        created_epoch = time.time()
        created_datetime = datetime.fromtimestamp(created_epoch).strftime(
            "%Y%m%d-%H%M%S"
        )

        self.runner_id = slugify(runner_id, lowercase=True)
        if self.runner_id != runner_id:
            raise RuntimeError(
                "[Session] Failed to initialise Session. Invalid Runner ID."
            )

        self.runner_args = runner_args
        self.runner_type = runner_type
        self.results_file_path = results_file_path
        self.progress_callback_func = progress_callback_func
        self.database_instance = database_instance

        self.red_teaming_progress = RedTeamingProgress(
            self.runner_id, self.runner_args, self.progress_callback_func
        )

        self.cancel_event = asyncio.Event()

        prompt_template = self.runner_args.get("prompt_template", "")
        context_strategy = self.runner_args.get("context_strategy", "")
        cs_num_of_prev_prompts = self.runner_args.get(
            "cs_num_of_prev_prompts", Session.DEFAULT_CONTEXT_STRATEGY_PROMPT
        )
        attack_module = self.runner_args.get("attack_module", "")
        system_prompt = self.runner_args.get("system_prompt", "")
        metric_id = self.runner_args.get("metric_id", "")

        self.check_file_exists(
            EnvVariables.PROMPT_TEMPLATES.name,
            prompt_template,
            "Prompt Template",
            "json",
        )
        self.check_file_exists(
            EnvVariables.CONTEXT_STRATEGY.name,
            context_strategy,
            "Context Strategy",
            "py",
        )
        self.check_file_exists(
            EnvVariables.ATTACK_MODULES.name, attack_module, "Attack Module", "py"
        )
        self.check_file_exists(EnvVariables.METRICS.name, metric_id, "Metric", "py")

        if self.database_instance:
            # create session metadata table if it does not exist
            if not Storage.check_database_table_exists(
                self.database_instance, "session_metadata_table"
            ):
                Storage.create_database_table(
                    self.database_instance, Session.sql_create_session_metadata_table
                )

            # get session metadata record
            session_metadata_records = Storage.read_database_records(
                self.database_instance, Session.sql_read_session_metadata
            )

            # check if the session metadata record already exists
            if session_metadata_records:
                logger.info("[Session] Session already exists.")
                self.session_metadata = SessionMetadata.from_tuple(
                    session_metadata_records[0]
                )
            # create a new record if session metadata does not exist
            else:
                logger.info("[Session] Creating new session.")

                # create chat history table for each endpoint

                self.session_metadata = SessionMetadata(
                    runner_id,
                    endpoints,
                    created_epoch,
                    created_datetime,
                    prompt_template,
                    context_strategy,
                    cs_num_of_prev_prompts,
                    attack_module,
                    metric_id,
                    system_prompt,
                )

                for endpoint in endpoints:
                    endpoint_id = endpoint.replace("-", "_")
                    Storage.create_database_table(
                        self.database_instance,
                        Session.sql_create_chat_history_table.format(endpoint_id),
                    )

                Storage.create_database_record(
                    self.database_instance,
                    self.session_metadata.to_tuple(),
                    Session.sql_create_session_metadata_record,
                )
        else:
            raise RuntimeError(
                "[Session] Failed to initialise Session. No database instance provided."
            )

    @staticmethod
    def load(database_instance: DBInterface | None) -> dict | None:
        """
        Loads run data for a given session_id from the database, or the latest run if run_id is None.

        This method retrieves run data for the specified run_id from the database, or if run_id is None,
        it retrieves the latest run. If the database instance is not provided, it raises a RuntimeError.
        If the database instance is provided, it invokes the read_record method of the database instance
        with the given run_id or the latest run and returns a RunArguments object created from the retrieved record.

        Parameters:
            database_instance (DBInterface | None): The database accessor instance.
            run_id (int | None): The ID of the run to retrieve, or None to retrieve the latest run.

        Returns:
            RunArguments: An object containing the details of the run with the given run_id or the latest run.
        """
        if not database_instance:
            raise RuntimeError("[Session] Runner instance database not provided.")

        # runner does not have session, return None
        if not Storage.check_database_table_exists(
            database_instance, "session_metadata_table"
        ):
            return None

        # retrieve session metadata
        session_metadata_info = Storage.read_database_records(
            database_instance,
            Session.sql_read_session_metadata,
        )

        if not session_metadata_info:
            raise RuntimeError("[Session] Failed to get Session metadata.")

        # convert session metadata from tuple to dict
        session_metadata_obj = SessionMetadata.from_tuple(session_metadata_info[0])
        session_metadata_dict = session_metadata_obj.to_dict()

        return session_metadata_dict

    async def run(self) -> list | None:
        """
        Asynchronously executes the session run process.

        This method orchestrates the entire session run process asynchronously. It initializes the session,
        sets up the necessary environment, executes the session's main logic, handles any errors, and finally,
        compiles and returns the results in a dictionary format. Throughout the process, it updates
        the session's status and logs progress.

        Returns:
            dict: A dictionary containing the results of the session run, including any errors encountered.
        """
        # ------------------------------------------------------------------------------
        # Part 1: Get asyncio running loop
        # ------------------------------------------------------------------------------
        logger.debug("[Session] Part 1: Loading asyncio running loop...")
        loop = asyncio.get_running_loop()

        # ------------------------------------------------------------------------------
        # Part 2: Load runner processing module
        # ------------------------------------------------------------------------------
        logger.debug("[Session] Part 2: Loading runner processing module...")
        start_time = time.perf_counter()
        runner_module_instance = None
        try:
            runner_processing_module_name = self.runner_args.get(
                "runner_processing_module", None
            )
            if runner_processing_module_name:
                # Intialize the runner instance
                runner_module_instance = get_instance(
                    runner_processing_module_name,
                    Storage.get_filepath(
                        EnvVariables.RUNNERS_MODULES.name,
                        runner_processing_module_name,
                        "py",
                    ),
                )
                if runner_module_instance:
                    runner_module_instance = runner_module_instance()
                else:
                    raise RuntimeError(
                        f"Unable to get defined runner module instance - {runner_module_instance}"
                    )
            else:
                raise RuntimeError(
                    f"Failed to get runner processing module name: {runner_processing_module_name}"
                )

        except Exception as e:
            logger.error(
                f"[Session] Failed to load runner processing module in Part 2 due to error: {str(e)}"
            )
            raise e

        finally:
            logger.debug(
                f"[Session] Loading runner processing module took {(time.perf_counter() - start_time):.4f}s"
            )

        # ------------------------------------------------------------------------------
        # Part 3: Run runner processing module
        # ------------------------------------------------------------------------------
        logger.debug("[Session] Part 3: Running runner processing module...")
        start_time = time.perf_counter()
        runner_results = {}

        try:
            if runner_module_instance:
                runner_results = await runner_module_instance.generate(  # type: ignore ; ducktyping
                    loop,
                    self.runner_args,
                    self.database_instance,
                    self.session_metadata,
                    self.check_redteaming_type(),
                    self.red_teaming_progress,
                    self.cancel_event,
                )
            else:
                raise RuntimeError("Failed to initialise runner module instance.")

        except Exception as e:
            logger.error(
                f"[Session] Failed to run runner processing module in Part 3 due to error: {str(e)}"
            )
            raise e

        finally:
            self.red_teaming_progress.status = RunStatus.COMPLETED
            if self.check_redteaming_type() == RedTeamingType.AUTOMATED:
                self.red_teaming_progress.notify_progress()
            logger.debug(
                f"[Session] Running runner processing module took {(time.perf_counter() - start_time):.4f}s"
            )

        # ------------------------------------------------------------------------------
        # Part 4: Wrap up run
        # ------------------------------------------------------------------------------
        logger.debug("[Session] Part 4: Wrap up run...")
        return runner_results

    def cancel(self) -> None:
        """
        Sets the cancel event to stop the automated red teaming process.

        This method is used to signal that the automated red teaming process should be cancelled. It sets the
        cancel_event which can be checked in various points of the asynchronous red teaming process to gracefully stop
        the execution.

        Returns:
            None
        """
        logger.warning("[Session] Cancelling automated red teaming...")
        self.cancel_event.set()

    def check_redteaming_type(self) -> RedTeamingType:
        """
        Checks the type of red teaming strategy based on the runner arguments.

        Returns:
            RedTeamingType: The type of red teaming strategy.

        Raises:
            RuntimeError: If the red teaming arguments are missing.
        """
        if (
            "attack_strategies" in self.runner_args
            and self.runner_args.get("attack_strategies") is not None
        ):
            return RedTeamingType.AUTOMATED
        elif (
            "manual_rt_args" in self.runner_args
            and self.runner_args.get("manual_rt_args") is not None
        ):
            return RedTeamingType.MANUAL
        else:
            raise RuntimeError("Missing red teaming arguments.")

    @staticmethod
    def update_context_strategy(
        db_instance: DBInterface | None, runner_id: str, context_strategy: str
    ) -> bool:
        """
        Updates the context strategy for a specific runner in the database.

        Args:
            db_instance (DBInterface | None): The database instance to update the context strategy in.
            runner_id (str): The ID of the runner.
            context_strategy (str): The name of the context strategy to be used.

        Returns:
            bool: The status on whether the context strategy is updated successfully.

        Raises:
            RuntimeError: If the database instance is not provided or if the context strategy does not exist.
        """
        if not db_instance:
            raise RuntimeError("[Session] Database instance not provided.")
        if context_strategy and not Storage.is_object_exists(
            EnvVariables.CONTEXT_STRATEGY.name, context_strategy, "py"
        ):
            raise RuntimeError(
                f"[Session] Context Strategy {context_strategy} does not exist."
            )
        else:
            Storage.update_database_record(
                db_instance,
                (context_strategy, runner_id),
                Session.sql_update_session_metadata_field.format("context_strategy"),
            )
            return True

    @staticmethod
    def update_cs_num_of_prev_prompts(
        db_instance: DBInterface | None, runner_id: str, cs_num_of_prev_prompts: int
    ) -> bool:
        """
        Updates the number of previous prompts for a specific runner in the database.

        Args:
            db_instance (DBInterface | None): The database instance to update the number of previous prompts in.
            runner_id (str): The ID of the runner.
            cs_num_of_prev_prompts (int): The new number of previous prompts to be used.

        Returns:
            bool: The status on whether the number of prompts for context strategy is updated successfully.

        Raises:
            RuntimeError: If the database instance is not provided.
        """
        if not db_instance:
            raise RuntimeError("[Session] Database instance not provided.")
        Storage.update_database_record(
            db_instance,
            (cs_num_of_prev_prompts, runner_id),
            Session.sql_update_session_metadata_field.format("cs_num_of_prev_prompts"),
        )
        return True

    @staticmethod
    def update_prompt_template(
        db_instance: DBInterface | None, runner_id: str, prompt_template: str
    ) -> bool:
        """
        Updates the prompt template in the database for the specified runner.

        Args:
            db_instance (DBInterface | None): The database instance to update the prompt template in.
            runner_id (str): The ID of the runner.
            prompt_template (str): The new prompt template to be used.

        Returns:
            bool: The status on whether the prompt template is updated successfully.

            Raises:
                RuntimeError: If the database instance is not provided or if the prompt template does not exist.
        """
        if not db_instance:
            raise RuntimeError("[Session] Database instance not provided.")
        if prompt_template and not Storage.is_object_exists(
            EnvVariables.PROMPT_TEMPLATES.name, prompt_template, "json"
        ):
            raise RuntimeError(
                f"[Session] Prompt Template {prompt_template} does not exist."
            )

        Storage.update_database_record(
            db_instance,
            (prompt_template, runner_id),
            Session.sql_update_session_metadata_field.format("prompt_template"),
        )
        return True

    @staticmethod
    def update_metric(
        db_instance: DBInterface | None, runner_id: str, metric_id: str
    ) -> bool:
        """
        Updates the metric in the database for the specified runner.

        Args:
            db_instance (DBInterface | None): The database instance to update the metric in.
            runner_id (str): The ID of the runner.
            metric_id (str): The new metric to be used.

        Returns:
            bool: The status on whether the metric is updated successfully.

        Raises:
            RuntimeError: If the database instance is not provided or if the metric does not exist.
        """
        if not db_instance:
            raise RuntimeError("[Session] Database instance not provided.")
        if metric_id and not Storage.is_object_exists(
            EnvVariables.METRICS.name, metric_id, "py"
        ):
            raise RuntimeError(f"[Session] Metric {metric_id} does not exist.")
        else:
            Storage.update_database_record(
                db_instance,
                (metric_id, runner_id),
                Session.sql_update_session_metadata_field.format("metric"),
            )
            return True

    @staticmethod
    def update_system_prompt(
        db_instance: DBInterface | None, runner_id: str, system_prompt: str
    ) -> bool:
        """
        Updates the system prompt in the database for the specified runner.

        Args:
            db_instance (DBInterface | None): The database instance to update the system prompt in.
            runner_id (str): The ID of the runner.
            system_prompt (str): The new system prompt to be used.

        Returns:
            bool: The status on whether the system prompt is updated successfully.

        Raises:
            RuntimeError: If the database instance is not provided.
        """
        if not db_instance:
            raise RuntimeError("[Session] Database instance not provided.")
        Storage.update_database_record(
            db_instance,
            (system_prompt, runner_id),
            Session.sql_update_session_metadata_field.format("system_prompt"),
        )
        return True

    @staticmethod
    def update_attack_module(
        db_instance: DBInterface | None, runner_id: str, attack_module_id: str
    ) -> bool:
        """
        Updates the attack module in the database for the specified runner.

        Args:
            db_instance (DBInterface | None): The database instance to update the attack module in.
            runner_id (str): The ID of the runner.
            attack_module_id (str): The new attack module to be used.

        Returns:
            bool: The status on whether the attack module is updated successfully.

        Raises:
            RuntimeError: If the database instance is not provided or if the attack module does not exist.
        """
        if not db_instance:
            raise RuntimeError("[Session] Database instance not provided.")
        if attack_module_id and not Storage.is_object_exists(
            EnvVariables.ATTACK_MODULES.name, attack_module_id, "py"
        ):
            raise RuntimeError(
                f"[Session] Attack Module {attack_module_id} does not exist."
            )
        else:
            Storage.update_database_record(
                db_instance,
                (attack_module_id, runner_id),
                Session.sql_update_session_metadata_field.format("attack_module"),
            )
            return True

    @staticmethod
    def delete(database_instance: DBInterface | None) -> bool:
        """
        Deletes the session metadata and associated endpoint tables from the database.

        Args:
            database_instance (DBInterface | None): The database instance to delete the session from.

        Returns:
            bool: The status on whether the session is deleted successfully.

        Raises:
            RuntimeError: If the database instance is not provided or if failed to get session metadata.
        """
        if not database_instance:
            raise RuntimeError("[Session] Database instance not provided.")

        session_metadata_info = Storage.read_database_records(
            database_instance,
            Session.sql_read_session_metadata,
        )
        if not session_metadata_info:
            raise RuntimeError("[Session] Failed to get Session metadata.")

        session_metadata_obj = SessionMetadata.from_tuple(session_metadata_info[0])
        Storage.delete_database_table(
            database_instance, Session.sql_drop_table.format("session_metadata_table")
        )
        for endpoint in session_metadata_obj.endpoints:
            endpoint = endpoint.replace("-", "_")
            Storage.delete_database_table(
                database_instance, Session.sql_drop_table.format(endpoint)
            )
        return True

    @staticmethod
    def get_session_chats(database_instance: DBInterface | None) -> dict:
        """
        Retrieves the chat history for all endpoints in a session.

        Args:
            database_instance (DBInterface | None): The database instance to retrieve the chat history from.

        Raises:
            RuntimeError: If the database instance is not provided.

        Returns:
            dict: A dictionary where the keys are endpoint IDs and the values are lists of chat history
            for each endpoint.
        """
        if not database_instance:
            raise RuntimeError("[Session] Database instance not provided.")

        session_metadata = Session.load(database_instance)
        chats = {}
        if session_metadata is not None and "endpoints" in session_metadata:
            endpoint_list = session_metadata.get("endpoints", [])
            for endpoint_id in endpoint_list:
                list_of_chats_from_one_ep = Chat.load_chat_history(
                    database_instance, endpoint_id.replace("-", "_")
                )
                chats.update({endpoint_id: list_of_chats_from_one_ep})
        return chats

    def check_file_exists(
        self, env_var_name: str, file_name: str, file_type: str, extension: str
    ) -> None:
        """
        Checks if a specified file exists in the storage.

        Args:
            env_var_name (str): The environment variable name where the file is stored.
            file_name (str): The name of the file to check.
            file_type (str): The type of the file.
            extension (str): The extension of the file.

        Raises:
            RuntimeError: If the file does not exist in the storage.
        """
        if file_name and not Storage.is_object_exists(
            env_var_name, file_name, extension
        ):
            raise RuntimeError(f"[Session] {file_type} {file_name} does not exist.")



import cmd2
from rich.console import Console
from rich.table import Table

from moonshot.api import (
    api_delete_runner,
    api_get_all_run,
    api_get_all_runner,
    api_get_available_session_info,
    api_load_session,
    api_read_runner,
)
from moonshot.integrations.cli.cli_errors import (
    ERROR_BENCHMARK_DELETE_RUNNER_RUNNER_VALIDATION,
    ERROR_BENCHMARK_VIEW_RUNNER_RUNNER_VALIDATION,
)
from moonshot.integrations.cli.common.display_helper import (
    display_view_list_format,
    display_view_str_format,
)

console = Console()


# ------------------------------------------------------------------------------
# CLI Functions
# ------------------------------------------------------------------------------
def list_runners() -> None:
    """
    List all runners.

    Retrieves and displays information about all runners, including their associated runs and session
    information. Fetches the data using the api_get_all_runner, api_get_all_run, and api_get_available_session_info
    functions, then calls the _display_runners function to present it in a user-friendly format.

    Returns:
        None
    """
    try:
        runner_info = api_get_all_runner()
        runner_run_info = api_get_all_run()
        _, runner_session_info = api_get_available_session_info()
        _display_runners(runner_info, runner_run_info, runner_session_info)
    except Exception as e:
        print(f"[list_runners]: {str(e)}")


def view_runner(args) -> None:
    """
    View a specific runner.

    Retrieves and displays information about a specific runner, including its associated runs and session
    information. Uses the runner identifier provided in the arguments to fetch the data and then calls the
    _display_runners function to present it in a user-friendly format.

    Args:
        args (argparse.Namespace): A namespace object from argparse. It should have the following attribute:
            runner (str): The identifier of the runner to view.

    Returns:
        None
    """
    try:
        if not isinstance(args.runner, str) or not args.runner or args.runner is None:
            raise TypeError(ERROR_BENCHMARK_VIEW_RUNNER_RUNNER_VALIDATION)

        runner_info = api_read_runner(args.runner)
        runner_run_info = api_get_all_run(args.runner)
        runner_session_info = api_load_session(args.runner)
        _display_runners([runner_info], runner_run_info, [runner_session_info])
    except Exception as e:
        print(f"[view_runner]: {str(e)}")


def delete_runner(args) -> None:
    """
    Delete a runner.

    Deletes a runner with the specified identifier. Prompts the user for confirmation before proceeding
    with the deletion. If the user confirms, it calls the api_delete_runner function from the moonshot.api module to
    delete the runner. If the deletion is successful, it prints a confirmation message. If an exception occurs, it
    prints an error message.

    Args:
        args (argparse.Namespace): A namespace object from argparse. It should have the following attribute:
            runner (str): The identifier of the runner to delete.

    Returns:
        None
    """
    # Confirm with the user before deleting a runner
    confirmation = console.input(
        "[bold red]Are you sure you want to delete the runner (y/N)? [/]"
    )
    if confirmation.lower() != "y":
        console.print("[bold yellow]Runner deletion cancelled.[/]")
        return

    try:
        if args.runner is None or not isinstance(args.runner, str) or not args.runner:
            raise ValueError(ERROR_BENCHMARK_DELETE_RUNNER_RUNNER_VALIDATION)

        api_delete_runner(args.runner)
        print("[delete_runner]: Runner deleted.")
    except Exception as e:
        print(f"[delete_runner]: {str(e)}")


# ------------------------------------------------------------------------------
# Helper functions: Display on cli
# ------------------------------------------------------------------------------
def _display_runners(
    runner_list: list, runner_run_info_list: list, runner_session_info_list: list
) -> None:
    """
    Display runners in a table format.

    Takes lists of runner information, run information, and session information, then displays them in a
    table format on the command line interface. Each runner is listed with details such as the runner's ID, name,
    description, number of runs, number of sessions, database file, and endpoints.

    Args:
        runner_list (list): A list of dictionaries, where each dictionary contains information about a runner.
        runner_run_info_list (list): A list of dictionaries, where each dictionary contains information about a run
            associated with a runner.
        runner_session_info_list (list): A list of dictionaries, where each dictionary contains information about a
            session associated with a runner.

    Returns:
        None
    """
    if runner_list:
        table = Table(
            title="List of Runners", show_lines=True, expand=True, header_style="bold"
        )
        table.add_column("No.", width=2)
        table.add_column("Runner", justify="left", width=78)
        table.add_column("Contains", justify="left", width=20, overflow="fold")
        for runner_id, runner in enumerate(runner_list, 1):
            (id, name, db_file, endpoints, description) = runner.values()

            db_info = display_view_str_format("Database", db_file)
            endpoints_info = display_view_list_format("Endpoints", endpoints)

            runs_count = sum(
                run_info["runner_id"] == id for run_info in runner_run_info_list
            )
            # Handle the case where session_info can be None
            sessions_count = sum(
                session_info is not None and session_info["session_id"] == id
                for session_info in runner_session_info_list
            )

            runner_info = (
                f"[red]id: {id}[/red]\n\n[blue]{name}[/blue]\n{description}\n"
                f"[blue]Number of Runs:[/blue] {runs_count}\n"
                f"[blue]Number of Sessions:[/blue] {sessions_count}"
            )
            contains_info = f"{db_info}\n\n{endpoints_info}"

            table.add_section()
            table.add_row(str(runner_id), runner_info, contains_info)
        console.print(table)
    else:
        console.print("[red]There are no runners found.[/red]")


# ------------------------------------------------------------------------------
# Cmd2 Arguments Parsers
# ------------------------------------------------------------------------------
# View runner arguments
view_runner_args = cmd2.Cmd2ArgumentParser(
    description="View a runner.",
    epilog="Example:\n view_runner my-new-cookbook-runner",
)
view_runner_args.add_argument("runner", type=str, help="Name of the runner")

# Delete runner arguments
delete_runner_args = cmd2.Cmd2ArgumentParser(
    description="Delete a runner.",
    epilog="Example:\n delete_runner my-new-cookbook-runner",
)
delete_runner_args.add_argument("runner", type=str, help="Name of the runner")



from moonshot.src.recipes.recipe_arguments import RecipeArguments as Recipe

from .... import api as moonshot_api
from ..schemas.recipe_create_dto import RecipeCreateDTO, RecipeUpdateDTO
from ..schemas.recipe_response_model import RecipeResponseModel
from ..services.base_service import BaseService
from ..services.utils.exceptions_handler import exception_handler


class RecipeService(BaseService):
    @exception_handler
    def create_recipe(self, recipe_data: RecipeCreateDTO) -> None:
        """
        Create a new recipe with the given data.

        Args:
            recipe_data (RecipeCreateDTO): Data transfer object containing recipe details.
        """
        moonshot_api.api_create_recipe(
            name=recipe_data.name,
            description=recipe_data.description,
            tags=recipe_data.tags,
            categories=recipe_data.categories,
            datasets=recipe_data.datasets,
            prompt_templates=recipe_data.prompt_templates,
            metrics=recipe_data.metrics,
            grading_scale=recipe_data.grading_scale,
        )

    @exception_handler
    def get_all_recipes(
        self,
        tags: str,
        categories: str,
        sort_by: str,
        count: bool,
        ids: str | None = None,
    ) -> list[RecipeResponseModel]:
        """
        Retrieve all recipes, with optional filters for tags, categories, sorting, and including prompt counts.

        Args:
            ids (str, optional): Filter recipes by IDs. If None, no ID-based filtering is applied.
            tags (str, optional): Filter recipes by tags. If None, no tag-based filtering is applied.
            categories (str, optional): Filter recipes by categories. If None, no category-based filtering is applied.
            sort_by (str, optional): Sort recipes by a specified field. If None, no sorting is applied.
            count (bool, optional): Include the total prompt count in each recipe if True.

        Returns:
            list[RecipeResponseModel]: A list of recipe, filtered and sorted, with optional prompt counts.
        """
        filtered_recipes: list[RecipeResponseModel] = []

        if ids:
            recipe_ids = ids.split(",")
            recipes = [moonshot_api.api_read_recipe(id) for id in recipe_ids]
        else:
            recipes = moonshot_api.api_get_all_recipe()

        for recipe_dict in recipes:
            recipe = RecipeResponseModel(**recipe_dict)
            if count:
                recipe.total_prompt_in_recipe, _ = get_total_prompt_in_recipe(recipe)
            filtered_recipes.append(recipe)

        # TODO - do all filtering in 1 pass
        if tags:
            filtered_recipes = [
                recipe for recipe in filtered_recipes if tags in recipe.tags
            ]

        if categories:
            categories_list = categories.split(",") if categories else []
            if categories_list:
                filtered_recipes = [
                    recipe
                    for recipe in filtered_recipes
                    if any(
                        category.lower() in (cat.lower() for cat in recipe.categories)
                        for category in categories_list
                    )
                ]
        if sort_by:
            if sort_by == "id":
                filtered_recipes.sort(key=lambda x: x.id)

        for recipe in filtered_recipes:
            recipe.required_config = get_metric_dependency_in_recipe(recipe)

        return filtered_recipes

    @exception_handler
    def get_all_recipes_name(self) -> list[str]:
        """
        Retrieve the names of all recipes.

        Returns:
            list[str]: A list of recipe names.
        """
        recipes = moonshot_api.api_get_all_recipe_name()
        return recipes

    @exception_handler
    def update_recipe(self, recipe_data: RecipeUpdateDTO, recipe_id: str) -> None:
        """
        Update an existing recipe with new data.

        Args:
            recipe_data (RecipeCreateDTO): Data transfer object containing new recipe details.
            recipe_id (str): The ID of the recipe to update.
        """
        update_data = {
            k: v
            for k, v in recipe_data.to_dict().items()
            if v is not None and k not in ["id", "stats"]
        }

        moonshot_api.api_update_recipe(rec_id=recipe_id, **update_data)

    @exception_handler
    def delete_recipe(self, recipe_id: str) -> None:
        """
        Delete a recipe by its ID.

        Args:
            recipe_id (str): The ID of the recipe to delete.
        """
        moonshot_api.api_delete_recipe(recipe_id)


@staticmethod
def get_total_prompt_in_recipe(recipe: Recipe) -> tuple[int, int]:
    """
    Calculate the total number of prompts in a recipe.

    This function sums up the number of dataset prompts and then multiplies
    the result by the number of prompt templates if they exist.

    Args:
        recipe (Recipe): The recipe object containing the stats and
                                  prompt templates information.

    Returns:
        int: The total count of prompts within the recipe.
        int: The total count of datasets within the recipe.
    """
    # Initialize total prompt count
    total_prompt_count = 0

    # Add counts from dataset prompts if available
    datasets_prompts = recipe.stats.get("num_of_datasets_prompts", {})
    total_prompt_count += sum(datasets_prompts.values())

    # If there are prompt templates, scale the total count by the number of templates
    if recipe.prompt_templates:
        total_prompt_count *= len(recipe.prompt_templates)

    return total_prompt_count, int(recipe.stats.get("num_of_datasets", 0))


@staticmethod
def get_metric_dependency_in_recipe(recipe: Recipe) -> dict | None:
    """
    Retrieve endpoint and configuration dependencies for a given recipe.

    This function gathers all available metrics along with their endpoints and configurations,
    then identifies and compiles the dependencies based on the metrics present in the provided recipe.

    Args:
        recipe (Recipe): The recipe object containing metrics information.

    Returns:
        dict[str, dict[str, list[str]]] | None: A dictionary with 'endpoints' and 'configurations' as keys,
        where 'endpoints' is a list of endpoint dependencies and 'configurations' is a dictionary of configuration
        dependencies. Returns None if no dependencies are found.
    """
    metrics = recipe.metrics
    all_metrics = moonshot_api.api_get_all_metric()
    aggregated_endpoints = set()
    aggregated_configurations = {}

    for metric in metrics:
        metric_data = next((m for m in all_metrics if m["id"] == metric), None)
        if metric_data:
            # Aggregate endpoints
            aggregated_endpoints.update(metric_data.get("endpoints", []))

            # Aggregate configurations
            for key, value in metric_data.get("configurations", {}).items():
                if key in aggregated_configurations:
                    aggregated_configurations[key].extend(
                        v for v in value if v not in aggregated_configurations[key]
                    )
                else:
                    aggregated_configurations[key] = value

    aggregated_data = {
        "endpoints": list(aggregated_endpoints),
        "configurations": aggregated_configurations,
    }
    return (
        aggregated_data
        if aggregated_data["endpoints"] or aggregated_data["configurations"]
        else None
    )



from __future__ import annotations

from abc import abstractmethod
from typing import Any


class DBInterface:
    @abstractmethod
    def create_connection(self) -> Any | None:
        """
        This method is used to create a connection to the database. The details of the connection
        are implementation specific and should be provided by the concrete class that inherits from this abstract class.

        Returns:
            Any | None: The connection object if successful, None otherwise.
        """
        pass

    @abstractmethod
    def close_connection(self) -> None:
        """
        This method is used to close the connection to the database. The details of the operation
        are implementation specific and should be provided by the concrete class that inherits from this abstract class.
        """
        pass

    @abstractmethod
    def create_table(self, create_table_sql: str) -> None:
        """
        This method is used to create a table in the database. The details of the operation
        are implementation specific and should be provided by the concrete class that inherits from this abstract class.

        Args:
            create_table_sql (str): The SQL query to create a table.

        Returns:
            None
        """
        pass

    @abstractmethod
    def create_record(self, record: tuple, create_record_sql: str) -> None:
        """
        This method is used to create a record in the database. The details of the operation
        are implementation specific and should be provided by the concrete class that inherits from this abstract class.

        Args:
            record (tuple): The record to be created.
            create_record_sql (str): The SQL query to create a record.

        Returns:
            None
        """
        pass

    @abstractmethod
    def read_record(self, record: tuple, read_record_sql: str) -> tuple | None:
        """
        This method is used to read a record from the database. The details of the operation
        are implementation specific and should be provided by the concrete class that inherits from this abstract class.

        Args:
            record (tuple): The record to be read.
            read_record_sql (str): The SQL query to read a record.

        Returns:
            tuple | None: The record if found, None otherwise.
        """
        pass

    @abstractmethod
    def read_records(self, read_records_sql: str) -> list[tuple] | None:
        """
        Executes a SQL query to read data from a table and returns the results.

        This method attempts to execute a provided SQL query to read data from a table within the SQLite database.
        If the connection to the database is established, it executes the query and returns all fetched rows as a list.
        In case of an error during the execution of the query, it prints an error message detailing the issue.

        Args:
            read_records_sql (str): The SQL query string used to read data from a table.

        Returns:
            list | None: A list of tuples representing the rows fetched by the query if successful.
        """
        pass

    @abstractmethod
    def update_record(self, record: tuple, update_record_sql: str) -> None:
        """
        This method is used to update a record in the database. The details of the operation
        are implementation specific and should be provided by the concrete class that inherits from this abstract class.

        Args:
            record (tuple): The record to be updated.
            update_record_sql (str): The SQL query to update a record.

        Returns:
            None
        """
        pass

    @abstractmethod
    def delete_records_in_table(self, delete_record_sql: str) -> None:
        """
        This method is used to delete all records from a specified table in the database. The details of the operation
        are implementation specific and should be provided by the concrete class that inherits from this abstract class.

        Args:
            delete_record_sql (str): The SQL query to delete all records from a table.

        Returns:
            None
        """
        pass

    @abstractmethod
    def check_database_table_exists(self, table_name: str) -> bool:
        """
        This method is used to check if a table exists in the database.

        Args:
            table_name (str): The name of the table to check for existence.

        Returns:
            bool: True if the table exists, False otherwise.
        """
        pass

    @abstractmethod
    def delete_database_table(self, delete_table_sql: str) -> None:
        """
        This method is used to delete a table from the database. The details of the operation
        are implementation specific and should be provided by the concrete class that inherits from this abstract class.

        Args:
            table_name (str): The name of the table to be deleted.

        Returns:
            None
        """
        pass



from __future__ import annotations

from typing import Annotated, Any

from pydantic import BaseModel, Field

from moonshot.src.connectors.connector_response import ConnectorResponse


class ConnectorPromptArguments(BaseModel):
    class Config:
        arbitrary_types_allowed = True

    prompt_index: Annotated[
        int, Field(strict=True, ge=0)
    ]  # The index of the prompt in the dataset

    prompt: str  # The actual prompt text

    target: Any  # The target response for the prompt

    predicted_results: ConnectorResponse | None = (
        None  # The predicted results, default is None
    )

    # The duration it took to get the results, must be a positive float
    duration: Annotated[float, Field(strict=True, ge=0.0)] = 0.0



from pydantic import validate_call

from moonshot.src.redteaming.attack.attack_module import AttackModule


# ------------------------------------------------------------------------------
# Red teaming APIs
# ------------------------------------------------------------------------------
def api_get_all_attack_modules() -> list[str]:
    """
    Retrieves all available attack module IDs.

    This function calls the `get_available_items` method from the `AttackModule` class to retrieve all available
    attack modules. It then extracts the IDs of each attack module and returns a list of these IDs.

    Returns:
        list[str]: A list of strings, each representing an attack module ID.
    """
    attack_modules_ids, _ = AttackModule.get_available_items()
    return attack_modules_ids


def api_get_all_attack_module_metadata() -> list[dict]:
    """
    Retrieves metadata for all available attack modules.

    This function calls the `get_available_items` method from the `AttackModule` class to retrieve all available
    attack modules. It then extracts the metadata for each attack module and returns a list of dictionaries,
    each containing the metadata of an attack module.

    Returns:
        list[dict]: A list of dictionaries, each representing the metadata of an attack module.
    """
    _, attack_modules_metadata = AttackModule.get_available_items()
    return attack_modules_metadata


@validate_call
def api_delete_attack_module(am_id: str) -> bool:
    """
    Deletes an attack module by its identifier.

    This function takes an attack module ID as input and calls the `delete` method from the `AttackModule` class
    to remove the specified attack module from storage.

    Args:
        am_id (str): The unique identifier of the attack module to be deleted.

    Returns:
        bool: True if the attack module was successfully deleted.

    Raises:
        Exception: If the deletion process encounters an error.
    """
    return AttackModule.delete(am_id)
